"""
Dataclasses representing Transactions for library use, most often
when sending a transaction to Starknet.
They should be compliant with the latest Starknet version.
"""

import base64
import dataclasses
import gzip
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, TypeVar, Union

import marshmallow
import marshmallow_dataclass
from marshmallow import fields

from starknet_py.hash.address import compute_address
from starknet_py.hash.transaction import (
    CommonTransactionV3Fields,
    TransactionHashPrefix,
    compute_declare_transaction_hash,
    compute_declare_v2_transaction_hash,
    compute_declare_v3_transaction_hash,
    compute_deploy_account_transaction_hash,
    compute_deploy_account_v3_transaction_hash,
    compute_invoke_transaction_hash,
    compute_invoke_v3_transaction_hash,
)
from starknet_py.net.client_models import (
    DAMode,
    DeprecatedContractClass,
    ResourceBoundsMapping,
    SierraContractClass,
    TransactionType,
)
from starknet_py.net.schemas.common import Felt
from starknet_py.net.schemas.rpc.contract import (
    ContractClassSchema,
    SierraContractClassSchema,
)

# TODO (#1219):
#  consider unifying these classes with client_models
#  remove marshmallow logic if not needed


@dataclass(frozen=True)
class Transaction(ABC):
    """
    Starknet transaction base class.
    """

    version: int = field(metadata={"marshmallow_field": Felt()})

    @property
    @abstractmethod
    def type(self) -> TransactionType:
        """
        Returns the corresponding TransactionType enum.
        """

    @abstractmethod
    def calculate_hash(self, chain_id: int) -> int:
        """
        Calculates the transaction hash in the Starknet network - a unique identifier of the
        transaction. See :py:meth:`~starknet_py.hash.transaction.compute_transaction_hash` docstring for more details.
        """


@dataclass(frozen=True)
class AccountTransaction(Transaction, ABC):
    """
    Represents a transaction in the Starknet network that is originated from an action of an
    account.
    """

    signature: List[int] = field(
        metadata={"marshmallow_field": fields.List(fields.String())}
    )
    nonce: int = field(metadata={"marshmallow_field": Felt()})


# Used instead of Union[Invoke, Declare, DeployAccount]
TypeAccountTransaction = TypeVar("TypeAccountTransaction", bound=AccountTransaction)


@dataclass(frozen=True)
class _DeprecatedAccountTransaction(AccountTransaction, ABC):
    max_fee: int = field(metadata={"marshmallow_field": Felt()})


@dataclass(frozen=True)
class _AccountTransactionV3(AccountTransaction, ABC):
    resource_bounds: ResourceBoundsMapping
    nonce_data_availability_mode: DAMode = field(init=False, default=DAMode.L1)
    fee_data_availability_mode: DAMode = field(init=False, default=DAMode.L1)
    paymaster_data: List[int] = field(init=False, default_factory=list)
    tip: int

    def get_common_fields(
        self,
        tx_prefix: TransactionHashPrefix,
        address: int,
        chain_id: int,
    ) -> CommonTransactionV3Fields:
        common_fields = [f.name for f in dataclasses.fields(_AccountTransactionV3)]

        # this is a helper function in a process to compute transaction hash
        # therefore signature is not included at this point
        common_fields.remove("signature")

        common_fields_with_values = {
            field_name: getattr(self, field_name) for field_name in common_fields
        }

        return CommonTransactionV3Fields(
            tx_prefix=tx_prefix,
            address=address,
            chain_id=chain_id,
            **common_fields_with_values,
        )


@dataclass(frozen=True)
class DeclareV3(_AccountTransactionV3):
    """
    Represents a transaction in the Starknet network that is a version 3 declaration of a Starknet contract
    class. Supports only sierra compiled contracts.
    """

    sender_address: int
    compiled_class_hash: int
    contract_class: SierraContractClass
    account_deployment_data: List[int] = field(default_factory=list)

    @property
    def type(self) -> TransactionType:
        return TransactionType.DECLARE

    def calculate_hash(self, chain_id: int) -> int:
        return compute_declare_v3_transaction_hash(
            account_deployment_data=self.account_deployment_data,
            contract_class=self.contract_class,
            compiled_class_hash=self.compiled_class_hash,
            common_fields=self.get_common_fields(
                tx_prefix=TransactionHashPrefix.DECLARE,
                address=self.sender_address,
                chain_id=chain_id,
            ),
        )


@dataclass(frozen=True)
class DeclareV2(_DeprecatedAccountTransaction):
    """
    Represents a transaction in the Starknet network that is a version 2 declaration of a Starknet contract
    class. Supports only sierra compiled contracts.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use :py:class:`~starknet_py.net.models.transaction.DeclareV3` instead.
    """

    contract_class: SierraContractClass = field(
        metadata={"marshmallow_field": fields.Nested(SierraContractClassSchema())}
    )
    compiled_class_hash: int = field(metadata={"marshmallow_field": Felt()})
    sender_address: int = field(metadata={"marshmallow_field": Felt()})

    @property
    def type(self) -> TransactionType:
        return TransactionType.DECLARE

    def calculate_hash(self, chain_id: int) -> int:
        return compute_declare_v2_transaction_hash(
            contract_class=self.contract_class,
            compiled_class_hash=self.compiled_class_hash,
            chain_id=chain_id,
            sender_address=self.sender_address,
            max_fee=self.max_fee,
            version=self.version,
            nonce=self.nonce,
        )


# pylint: disable=line-too-long
@dataclass(frozen=True)
class DeclareV1(_DeprecatedAccountTransaction):
    """
    Based on https://docs.starknet.io/architecture-and-concepts/network-architecture/transactions/#transaction_versioning

    Represents a transaction in the Starknet network that is a declaration of a Starknet contract
    class.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use :py:class:`~starknet_py.net.models.transaction.DeclareV3` instead.
    """

    # The class to be declared, included for all methods involving execution (estimateFee, simulateTransactions)
    contract_class: DeprecatedContractClass = field(
        metadata={"marshmallow_field": fields.Nested(ContractClassSchema())}
    )
    # The address of the account contract sending the declaration transaction.
    sender_address: int = field(metadata={"marshmallow_field": Felt()})

    @property
    def type(self) -> TransactionType:
        return TransactionType.DECLARE

    @marshmallow.post_dump
    def post_dump(self, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        # Allowing **kwargs is needed here because marshmallow is passing additional parameters here
        # along with data, which we don't handle.
        # pylint: disable=unused-argument, no-self-use
        return compress_program(data)

    @marshmallow.pre_load
    def pre_load(self, data: Dict[str, Any], **kwargs) -> Dict[str, Any]:
        # pylint: disable=unused-argument, no-self-use
        return decompress_program(data)

    def calculate_hash(self, chain_id: int) -> int:
        """
        Calculates the transaction hash in the Starknet network.
        """
        return compute_declare_transaction_hash(
            contract_class=self.contract_class,
            chain_id=chain_id,
            sender_address=self.sender_address,
            max_fee=self.max_fee,
            version=self.version,
            nonce=self.nonce,
        )


# pylint: enable=line-too-long


@dataclass(frozen=True)
class DeployAccountV3(_AccountTransactionV3):
    """
    Represents a transaction in the Starknet network that is a version 3 deployment of a Starknet account
    contract.
    """

    class_hash: int
    contract_address_salt: int
    constructor_calldata: List[int]

    @property
    def type(self) -> TransactionType:
        return TransactionType.DEPLOY_ACCOUNT

    def calculate_hash(self, chain_id: int) -> int:
        contract_address = compute_address(
            salt=self.contract_address_salt,
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            deployer_address=0,
        )
        return compute_deploy_account_v3_transaction_hash(
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            contract_address_salt=self.contract_address_salt,
            common_fields=self.get_common_fields(
                tx_prefix=TransactionHashPrefix.DEPLOY_ACCOUNT,
                address=contract_address,
                chain_id=chain_id,
            ),
        )


@dataclass(frozen=True)
class DeployAccountV1(_DeprecatedAccountTransaction):
    """
    Represents a transaction in the Starknet network that is a deployment of a Starknet account
    contract.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use :py:class:`~starknet_py.net.models.transaction.DeployAccountV3` instead
    """

    class_hash: int = field(metadata={"marshmallow_field": Felt()})
    contract_address_salt: int = field(metadata={"marshmallow_field": Felt()})
    constructor_calldata: List[int] = field(
        metadata={"marshmallow_field": fields.List(fields.String())}
    )

    @property
    def type(self) -> TransactionType:
        return TransactionType.DEPLOY_ACCOUNT

    def calculate_hash(self, chain_id: int) -> int:
        """
        Calculates the transaction hash in the Starknet network.
        """
        contract_address = compute_address(
            salt=self.contract_address_salt,
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            deployer_address=0,
        )
        return compute_deploy_account_transaction_hash(
            version=self.version,
            contract_address=contract_address,
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            max_fee=self.max_fee,
            nonce=self.nonce,
            salt=self.contract_address_salt,
            chain_id=chain_id,
        )


@dataclass(frozen=True)
class InvokeV3(_AccountTransactionV3):
    """
    Represents a transaction in the Starknet network that is a version 3 invocation of a Cairo contract
    function.
    """

    calldata: List[int]
    sender_address: int
    account_deployment_data: List[int] = field(default_factory=list)

    @property
    def type(self) -> TransactionType:
        return TransactionType.INVOKE

    def calculate_hash(self, chain_id: int) -> int:
        return compute_invoke_v3_transaction_hash(
            account_deployment_data=self.account_deployment_data,
            calldata=self.calldata,
            common_fields=self.get_common_fields(
                tx_prefix=TransactionHashPrefix.INVOKE,
                address=self.sender_address,
                chain_id=chain_id,
            ),
        )


@dataclass(frozen=True)
class InvokeV1(_DeprecatedAccountTransaction):
    """
    Represents a transaction in the Starknet network that is an invocation of a Cairo contract
    function.

    .. deprecated:: 0.25.0
        This class is deprecated and will be removed in future versions.
        Use :py:class:`~starknet_py.net.models.transaction.InvokeV3` instead
    """

    sender_address: int = field(metadata={"marshmallow_field": Felt()})
    calldata: List[int] = field(
        metadata={"marshmallow_field": fields.List(fields.String())}
    )

    @property
    def type(self) -> TransactionType:
        return TransactionType.INVOKE

    def calculate_hash(self, chain_id: int) -> int:
        """
        Calculates the transaction hash in the Starknet network.
        """
        return compute_invoke_transaction_hash(
            version=self.version,
            sender_address=self.sender_address,
            calldata=self.calldata,
            max_fee=self.max_fee,
            chain_id=chain_id,
            nonce=self.nonce,
        )


Declare = Union[DeclareV1, DeclareV2, DeclareV3]
DeployAccount = Union[DeployAccountV1, DeployAccountV3]
Invoke = Union[InvokeV1, InvokeV3]

InvokeV1Schema = marshmallow_dataclass.class_schema(InvokeV1)
DeclareV1Schema = marshmallow_dataclass.class_schema(DeclareV1)
DeclareV2Schema = marshmallow_dataclass.class_schema(DeclareV2)
DeployAccountV1Schema = marshmallow_dataclass.class_schema(DeployAccountV1)


def compress_program(data: dict, program_name: str = "program") -> dict:
    program = data["contract_class"][program_name]
    compressed_program = json.dumps(program)
    compressed_program = gzip.compress(data=compressed_program.encode("ascii"))
    compressed_program = base64.b64encode(compressed_program)
    data["contract_class"][program_name] = compressed_program.decode("ascii")
    return data


def decompress_program(data: dict, program_name: str = "program") -> dict:
    compressed_program: str = data["contract_class"][program_name]
    program = base64.b64decode(compressed_program.encode("ascii"))
    program = gzip.decompress(data=program)
    program = json.loads(program.decode("ascii"))
    data["contract_class"][program_name] = program
    return data



# pylint: disable=unused-variable
import pytest

from starknet_py.common import create_sierra_compiled_contract
from starknet_py.contract import Contract
from starknet_py.net.account.account import Account
from starknet_py.net.client_models import (
    InvokeTransactionV3,
    ResourceBounds,
    ResourceBoundsMapping,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import DeclareV3, StarknetChainId
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract


def test_init():
    # docs-start: init
    contract = Contract(
        address=0x123,
        abi=[
            {
                "inputs": [{"name": "amount", "type": "felt"}],
                "name": "increase_balance",
                "outputs": [],
                "type": "function",
            },
        ],
        provider=Account(
            address=0x321,
            client=FullNodeClient(node_url="https://your.node.url"),
            key_pair=KeyPair(12, 34),
            chain=StarknetChainId.SEPOLIA,
        ),
        cairo_version=0,
    )
    # docs-end: init


@pytest.mark.asyncio
async def test_from_address(account, contract_address):
    # docs-start: from_address
    address = 1 or 0x1 or "0x1"
    # docs-end: from_address
    address = contract_address
    # docs-start: from_address
    contract = await Contract.from_address(address=address, provider=account)
    # or if the contract is a proxy (read more about resolving proxies in the `Guide`)
    config = True
    # docs-end: from_address
    config = False
    # docs-start: from_address
    contract = await Contract.from_address(
        address=address, provider=account, proxy_config=config
    )
    # docs-end: from_address


@pytest.mark.asyncio
async def test_declare_v3(account):
    contract = load_contract(contract_name="TestContract", version=ContractVersion.V2)
    # docs-start: declare_v3
    # here `contract` is a dict containing sierra and casm artifacts
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    declare_result = await Contract.declare_v3(
        account,
        compiled_contract=contract["sierra"],
        compiled_contract_casm=contract["casm"],
        resource_bounds=resource_bounds,
    )
    # docs-end: declare_v3
    await declare_result.wait_for_acceptance()

    assert isinstance(declare_result.declare_transaction, DeclareV3)
    assert isinstance(declare_result.hash, int)
    assert isinstance(declare_result.class_hash, int)
    assert declare_result.compiled_contract == contract["sierra"]


@pytest.mark.asyncio
async def test_deploy_contract_v3(account, hello_starknet_class_hash: int):
    compiled_contract = load_contract("HelloStarknet")["sierra"]
    # docs-start: deploy_contract_v3
    abi = create_sierra_compiled_contract(
        compiled_contract=compiled_contract
    ).parsed_abi
    # docs-end: deploy_contract_v3
    class_hash = hello_starknet_class_hash
    # docs-start: deploy_contract_v3
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    deploy_result = await Contract.deploy_contract_v3(
        class_hash=class_hash,
        account=account,
        abi=abi,
        resource_bounds=resource_bounds,
    )
    # docs-end: deploy_contract_v3
    await deploy_result.wait_for_acceptance()

    contract = deploy_result.deployed_contract
    assert isinstance(contract.address, int)
    assert len(contract.functions) != 0

    transaction = await account.client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, InvokeTransactionV3)

    class_hash = await account.client.get_class_hash_at(
        contract_address=contract.address
    )
    assert class_hash == hello_starknet_class_hash


@pytest.mark.asyncio
async def test_deploy_contract_v3_without_abi(account, hello_starknet_class_hash: int):
    deploy_result = await Contract.deploy_contract_v3(
        class_hash=hello_starknet_class_hash,
        account=account,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await deploy_result.wait_for_acceptance()

    contract = deploy_result.deployed_contract
    assert isinstance(contract.address, int)
    assert len(contract.functions) == 2

    transaction = await account.client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, InvokeTransactionV3)

    class_hash = await account.client.get_class_hash_at(
        contract_address=contract.address
    )
    assert class_hash == hello_starknet_class_hash



import math
import statistics
from typing import Optional, Union

from starknet_py.net.client import Client
from starknet_py.net.client_models import Hash, Tag, TransactionV3


async def estimate_tip(
    client: Client,
    block_hash: Optional[Union[Hash, Tag]] = None,
    block_number: Optional[Union[int, Tag]] = None,
) -> int:
    """
    Estimates the transaction tip by taking the median of all V3 transaction tips in the specified block.
    If no block is provided, the `latest` block is used.

    :param client: Client instance.
    :param block_hash: Block's hash or literals `"l1_accepted"`, `"latest" or "pre_confirmed"`
    :param block_number: Block's number or literals `"l1_accepted"`, `"latest" or "pre_confirmed"``
    """
    if block_hash is None and block_number is None:
        block_hash = "latest"

    block_with_txs = await client.get_block_with_txs(block_hash, block_number)

    tips = [
        tx.tip for tx in block_with_txs.transactions if isinstance(tx, TransactionV3)
    ]

    if len(tips) == 0:
        return 0

    tips_median = statistics.median(tips)
    return math.ceil(tips_median)



# pylint: disable=unused-variable
import pytest

from starknet_py.contract import Contract


@pytest.mark.asyncio
async def test_call_raw(map_contract: Contract):
    prepared_function_call = map_contract.functions["get"].prepare_call(key=10)
    # docs-start: call_raw
    raw_response = await prepared_function_call.call_raw(block_number="latest")
    # or
    raw_response = await prepared_function_call.call_raw()
    # docs-end: call_raw


@pytest.mark.asyncio
async def test_call(map_contract: Contract):
    prepared_function_call = map_contract.functions["get"].prepare_call(key=10)
    # docs-start: call
    response = await prepared_function_call.call(block_number="latest")
    # or
    response = await prepared_function_call.call()
    # docs-end: call



import sys

import pytest

from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_declaring_contracts(
    account, map_compiled_contract_and_class_hash_copy_1
):
    (compiled_contract, class_hash) = map_compiled_contract_and_class_hash_copy_1

    # docs: start
    # Account.sign_declare_v3 takes a string containing a compiled contract (sierra)
    # and a class hash (casm_class_hash)

    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    declare_transaction = await account.sign_declare_v3(
        compiled_contract=compiled_contract,
        compiled_class_hash=class_hash,
        resource_bounds=resource_bounds,
    )

    # To declare a contract, send Declare transaction with Client.declare method
    resp = await account.client.declare(transaction=declare_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    declared_contract_class_hash = resp.class_hash
    # docs: end

    assert declared_contract_class_hash != 0



from marshmallow import Schema, fields
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from starknet_py.abi.v0.shape import (
    CONSTRUCTOR_ENTRY,
    EVENT_ENTRY,
    FUNCTION_ENTRY,
    L1_HANDLER_ENTRY,
    STRUCT_ENTRY,
)


class TypedParameterSchema(Schema):
    name = fields.String(data_key="name", required=True)
    type = fields.String(data_key="type", required=True)


class StructMemberSchema(TypedParameterSchema):
    offset = fields.Integer(data_key="offset", required=False)


class FunctionBaseSchema(Schema):
    name = fields.String(data_key="name", required=True)
    inputs = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="inputs", required=True
    )
    outputs = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="outputs", required=True
    )


class FunctionAbiEntrySchema(FunctionBaseSchema):
    type = fields.Constant(FUNCTION_ENTRY, data_key="type", required=True)


class ConstructorAbiEntrySchema(FunctionBaseSchema):
    type = fields.Constant(CONSTRUCTOR_ENTRY, data_key="type", required=True)


class L1HandlerAbiEntrySchema(FunctionBaseSchema):
    type = fields.Constant(L1_HANDLER_ENTRY, data_key="type", required=True)


class EventAbiEntrySchema(Schema):
    type = fields.Constant(EVENT_ENTRY, data_key="type", required=True)
    name = fields.String(data_key="name", required=True)
    keys = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="keys", required=True
    )
    data = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="data", required=True
    )


class StructAbiEntrySchema(Schema):
    type = fields.Constant(STRUCT_ENTRY, data_key="type", required=True)
    name = fields.String(data_key="name", required=True)
    size = fields.Integer(data_key="size", required=True)
    members = fields.List(
        fields.Nested(StructMemberSchema()), data_key="members", required=True
    )


class ContractAbiEntrySchema(OneOfSchema):
    type_field_remove = False
    type_schemas = {
        FUNCTION_ENTRY: FunctionAbiEntrySchema,
        L1_HANDLER_ENTRY: L1HandlerAbiEntrySchema,
        CONSTRUCTOR_ENTRY: ConstructorAbiEntrySchema,
        EVENT_ENTRY: EventAbiEntrySchema,
        STRUCT_ENTRY: StructAbiEntrySchema,
    }



from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.net.tip import estimate_tip
from starknet_py.tests.e2e.fixtures.misc import (
    starknet_block_mock,
    transaction_mock_with_tip,
)


@pytest.mark.asyncio
async def test_tip(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = [
            transaction_mock_with_tip(10),
            transaction_mock_with_tip(20),
            transaction_mock_with_tip(30),
        ]
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 20


@pytest.mark.asyncio
async def test_tip_no_txs(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = []
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 0


@pytest.mark.asyncio
async def test_tip_all_equal(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = [
            transaction_mock_with_tip(10),
            transaction_mock_with_tip(10),
            transaction_mock_with_tip(10),
        ]
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 10


@pytest.mark.asyncio
async def test_tip_even(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = [
            transaction_mock_with_tip(10),
            transaction_mock_with_tip(20),
            transaction_mock_with_tip(30),
            transaction_mock_with_tip(40),
        ]
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 25


@pytest.mark.asyncio
async def test_tip_zeroes(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = [
            transaction_mock_with_tip(0),
            transaction_mock_with_tip(0),
            transaction_mock_with_tip(30),
            transaction_mock_with_tip(40),
        ]
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 15


@pytest.mark.asyncio
async def test_tip_all_zeroes(get_block_with_txs_path, client):
    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        block = starknet_block_mock()
        block.transactions = [
            transaction_mock_with_tip(0),
            transaction_mock_with_tip(0),
        ]
        get_block_with_txs_mock.return_value = block

        assert await estimate_tip(client) == 0



import sys

import pytest

from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS

# docs-abi: start
abi = [
    {
        "inputs": [
            {"name": "sender", "type": "felt"},
            {"name": "recipient", "type": "felt"},
            {"name": "amount", "type": "felt"},
        ],
        "name": "transferFrom",
        "outputs": [{"name": "success", "type": "felt"}],
        "type": "function",
    },
    {
        "inputs": [{"name": "account", "type": "felt"}],
        "name": "balanceOf",
        "outputs": [{"name": "balance", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
]
# docs-abi: end


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_using_existing_contracts(account, erc20_contract):
    # pylint: disable=import-outside-toplevel,too-many-locals,unused-variable
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

    address = "0x00178130dd6286a9a0e031e4c73b2bd04ffa92804264a25c1c08c1612559f458"

    # When ABI is known statically just use the Contract constructor
    contract = Contract(address=address, abi=abi, provider=account, cairo_version=0)
    # or if it is not known
    # Contract.from_address makes additional request to fetch the ABI
    # docs: end

    address = erc20_contract.address
    # docs: start
    contract = await Contract.from_address(provider=account, address=address)

    sender = account.address
    recipient = "123"
    # docs: end
    invocation = await contract.functions["increase_allowance"].invoke_v3(
        sender,
        200,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await invocation.wait_for_acceptance()
    # docs: start

    # Using only positional arguments
    invocation = await contract.functions["transfer_from"].invoke_v3(
        sender,
        recipient,
        50,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    # docs: end
    await invocation.wait_for_acceptance()
    # docs: start

    # Using only keyword arguments
    invocation = await contract.functions["transfer_from"].invoke_v3(
        sender=sender,
        recipient=recipient,
        amount=50,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    # docs: end
    await invocation.wait_for_acceptance()
    # docs: start

    # Mixing positional with keyword arguments
    invocation = await contract.functions["transfer_from"].invoke_v3(
        sender,
        recipient,
        amount=50,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    # docs: end
    await invocation.wait_for_acceptance()
    # docs: start

    # Creating a prepared function call with arguments
    # It is also possible to use `prepare_invoke_v3`
    transfer = contract.functions["transfer_from"].prepare_invoke_v3(
        sender,
        recipient,
        amount=50,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    invocation = await transfer.invoke()

    # Wait for tx
    await invocation.wait_for_acceptance()

    (balance,) = await contract.functions["balance_of"].call(recipient)
    # docs: end

    assert balance == 200


@pytest.mark.asyncio
async def test_raw_call(account):
    # pylint: disable=import-outside-toplevel
    # docs-raw-call: start
    from starknet_py.hash.selector import get_selector_from_name
    from starknet_py.net.client_models import Call

    # docs-raw-call: end
    # fmt: off
    # docs-raw-call: start
    eth_token_address = 0x049D36570D4E46F48E99674BD3FCC84644DDD6B96F7C741B1562B82F9E004DC7
    # docs-raw-call: end
    # fmt: on
    # docs-raw-call: start

    # Create a call to function "balanceOf" at address `eth_token_address`
    call = Call(
        to_addr=eth_token_address,
        selector=get_selector_from_name("balanceOf"),
        calldata=[account.address],
    )
    # Pass the created call to Client.call_contract
    account_balance = await account.client.call_contract(call)

    # Note that a Contract instance cannot be used here, since it needs ABI to generate the functions
    # docs-raw-call: end

    assert account_balance != [0, 0]



from typing import Optional, Union

import pytest

from starknet_py.net.client_models import Call, Hash, Tag
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_custom_nonce(account):
    # pylint: disable=import-outside-toplevel
    client = account.client
    address = account.address
    private_key = account.signer.key_pair.private_key

    # docs: start
    from starknet_py.net.account.account import Account
    from starknet_py.net.client import Client
    from starknet_py.net.models import AddressRepresentation, StarknetChainId
    from starknet_py.net.signer import BaseSigner
    from starknet_py.net.signer.key_pair import KeyPair

    class MyAccount(Account):
        def __init__(
            self,
            *,
            address: AddressRepresentation,
            client: Client,
            signer: Optional[BaseSigner] = None,
            key_pair: Optional[KeyPair] = None,
            chain: Optional[StarknetChainId] = None,
        ):
            super().__init__(
                address=address,
                client=client,
                signer=signer,
                key_pair=key_pair,
                chain=chain,
            )
            # Create a simple counter that will store a nonce
            self.nonce_counter = 0

        async def get_nonce(
            self,
            *,
            block_hash: Optional[Union[Hash, Tag]] = None,
            block_number: Optional[Union[int, Tag]] = None,
        ) -> int:
            # Increment the counter and return the nonce.
            # This is just an example custom nonce logic and is not meant
            # to be a recommended solution.
            nonce = self.nonce_counter
            self.nonce_counter += 1
            return nonce

    account = MyAccount(
        address=address,
        client=client,
        key_pair=KeyPair.from_private_key(private_key),
        chain=StarknetChainId.SEPOLIA,
    )
    # docs: end

    assert account.nonce_counter == 0
    await account.sign_invoke_v3(
        calls=Call(0x1, 0x1, []), resource_bounds=MAX_RESOURCE_BOUNDS
    )
    assert account.nonce_counter == 1



import sys

import pytest

from starknet_py.contract import Contract
from starknet_py.net.client_errors import ClientError
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_impersonate_account(
    devnet_client_fork_mode, account_to_impersonate, f_string_contract
):
    await devnet_client_fork_mode.impersonate_account(
        address=account_to_impersonate.address
    )

    contract = await Contract.from_address(
        provider=account_to_impersonate, address=f_string_contract.address
    )

    invocation = await contract.functions["set_string"].invoke_v3(
        "test", resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await devnet_client_fork_mode.stop_impersonate_account(
        address=account_to_impersonate.address
    )

    assert (
        invocation.invoke_transaction.sender_address == account_to_impersonate.address
    )


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_auto_impersonate(
    devnet_client_fork_mode, account_to_impersonate, f_string_contract
):
    await devnet_client_fork_mode.auto_impersonate()

    contract = await Contract.from_address(
        provider=account_to_impersonate, address=f_string_contract.address
    )

    invocation = await contract.functions["set_string"].invoke_v3(
        "test", resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await devnet_client_fork_mode.stop_auto_impersonate()

    assert (
        invocation.invoke_transaction.sender_address == account_to_impersonate.address
    )


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_impersonated_account_should_fail(
    account_to_impersonate, f_string_contract
):
    contract = await Contract.from_address(
        provider=account_to_impersonate, address=f_string_contract.address
    )

    with pytest.raises(ClientError):
        await contract.functions["set_string"].invoke_v3("test", auto_estimate=True)



import asyncio

import pytest

from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class Function:
    def __init__(self):
        self.name = "function X"

    @staticmethod
    async def call():
        await asyncio.sleep(0.1)
        return 1

    @staticmethod
    async def failure():
        # pylint: disable=broad-exception-raised
        raise Exception("Error")

    def get_name(self):
        return self.name

    @staticmethod
    async def method():
        return 1

    @staticmethod
    def method_sync():
        # Shouldn't be overridden
        return 2


@add_sync_methods
class Repository:
    def __init__(self):
        self.function = Function()

    async def get_function(self):
        await asyncio.sleep(0.1)
        return self.function


# pylint: disable=no-member
@add_sync_methods
class Contract:
    def __init__(self, address):
        self.address = address

    @staticmethod
    async def get_repository():
        await asyncio.sleep(0.1)
        return Repository()

    @classmethod
    async def example_class_method(cls):
        await asyncio.sleep(0.1)
        return 2


@pytest.mark.asyncio
async def test_asynchronous_versions():
    contract = Contract("1")
    repository = await contract.get_repository()
    function = await repository.get_function()

    assert await function.call() == 1
    assert function.get_name() == "function X"
    assert await contract.example_class_method() == 2
    assert await Function.method() == 1
    assert Function.method_sync() == 2

    with pytest.raises(Exception):
        await function.failure()


def test_sync_versions():
    contract = Contract("1")
    # Ignore typing, because _sync is added dynamically
    repository = contract.get_repository_sync()  # pyright: ignore
    function = repository.get_function_sync()  # pyright: ignore

    assert function.call_sync() == 1
    assert function.get_name() == "function X"
    assert contract.example_class_method_sync() == 2  # pyright: ignore

    with pytest.raises(Exception):
        function.failure_sync()



from typing import Generator

import pytest


@pytest.fixture(scope="package")
def devnet_ws(devnet: str) -> Generator[str, None, None]:
    """
    Returns WebSocket address of devnet.
    """
    yield devnet.replace("http", "ws") + "/ws"



import pytest

from starknet_py.devnet_utils.devnet_client import DevnetClient


@pytest.fixture(name="devnet_client", scope="package")
def create_devnet_client(devnet) -> DevnetClient:
    return DevnetClient(node_url=devnet + "/rpc")


@pytest.fixture(name="devnet_client_fork_mode", scope="package")
def create_devnet_client_fork_mode(devnet_forking_mode) -> DevnetClient:
    return DevnetClient(node_url=devnet_forking_mode + "/rpc")



import pytest

from starknet_py.hash.address import (
    compute_address,
    get_checksum_address,
    is_checksum_address,
)


def test_compute_address():
    assert (
        compute_address(
            class_hash=951442054899045155353616354734460058868858519055082696003992725251069061570,
            constructor_calldata=[21, 37],
            salt=1111,
        )
        == 1357105550695717639826158786311415599375114169232402161465584707209611368775
    )


def test_compute_address_with_deployer_address():
    assert (
        compute_address(
            class_hash=951442054899045155353616354734460058868858519055082696003992725251069061570,
            constructor_calldata=[21, 37],
            salt=1111,
            deployer_address=1234,
        )
        == 3179899882984850239687045389724311807765146621017486664543269641150383510696
    )


@pytest.mark.parametrize(
    "address, checksum_address",
    [
        (
            "0x2fd23d9182193775423497fc0c472e156c57c69e4089a1967fb288a2d84e914",
            "0x02Fd23d9182193775423497fc0c472E156C57C69E4089A1967fb288A2d84e914",
        ),
        (
            "0x00abcdefabcdefabcdefabcdefabcdefabcdefabcdefabcdefabcdefabcdefab",
            "0x00AbcDefaBcdefabCDEfAbCDEfAbcdEFAbCDEfabCDefaBCdEFaBcDeFaBcDefAb",
        ),
        (
            "0xfedcbafedcbafedcbafedcbafedcbafedcbafedcbafedcbafedcbafedcbafe",
            "0x00fEdCBafEdcbafEDCbAFedCBAFeDCbafEdCBAfeDcbaFeDCbAfEDCbAfeDcbAFE",
        ),
        ("0xa", "0x000000000000000000000000000000000000000000000000000000000000000A"),
        (
            "0x0",
            "0x0000000000000000000000000000000000000000000000000000000000000000",
        ),
    ],
)
def test_get_checksum_address(address, checksum_address):
    assert get_checksum_address(address) == checksum_address


@pytest.mark.parametrize("address", ["", "0xx", "0123"])
def test_get_checksum_address_raises_on_invalid_address(address):
    with pytest.raises(ValueError):
        get_checksum_address(address)


@pytest.mark.parametrize(
    "address, is_checksum",
    [
        ("0x02Fd23d9182193775423497fc0c472E156C57C69E4089A1967fb288A2d84e914", True),
        ("0x000000000000000000000000000000000000000000000000000000000000000a", False),
    ],
)
def test_is_checksum_address(address, is_checksum):
    assert is_checksum_address(address) == is_checksum



"""
Dataclasses representing responses from Starknet Websocket RPC API.
"""

from dataclasses import dataclass
from typing import Generic, TypeVar

from starknet_py.net.client_models import (
    BlockHeader,
    EmittedEventWithFinalityStatus,
    Transaction,
    TransactionReceiptWithBlockInfo,
    TransactionStatusResponse,
    TransactionStatusWithoutL1,
)

T = TypeVar("T")


@dataclass
class Notification(Generic[T]):
    """
    Base class for notification.
    """

    subscription_id: str
    result: T


@dataclass
class NewHeadsNotification(Notification[BlockHeader]):
    """
    Notification to the client of a new block header.
    """


@dataclass
class NewEventsNotification(Notification[EmittedEventWithFinalityStatus]):
    """
    Notification to the client of a new event.
    """


@dataclass
class NewTransactionStatus:
    """
    New transaction status.
    """

    transaction_hash: int
    status: TransactionStatusResponse


@dataclass
class TransactionStatusNotification(Notification[NewTransactionStatus]):
    """
    Notification to the client of a new transaction status.
    """


@dataclass
class ReorgData:
    """
    Data about reorganized blocks, starting and ending block number and hash.
    """

    starting_block_hash: int
    starting_block_number: int
    ending_block_hash: int
    ending_block_number: int


@dataclass
class ReorgNotification(Notification[ReorgData]):
    """
    Notification of a reorganization of the chain.
    """


@dataclass
class NewTransactionReceiptsNotification(Notification[TransactionReceiptWithBlockInfo]):
    """
    Notification of a new transaction receipt
    """


@dataclass
class NewTransactionNotificationResult:
    transaction: Transaction
    finality_status: TransactionStatusWithoutL1


@dataclass
class NewTransactionNotification(Notification[NewTransactionNotificationResult]):
    """
    Notification of a new transaction
    """



import pytest

from starknet_py.net.account.account import Account
from starknet_py.net.models import StarknetChainId
from starknet_py.net.models.transaction import DeployAccountV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_general_flow(client, deploy_account_details_factory):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    account = Account(
        address=address,
        client=client,
        key_pair=key_pair,
        chain=StarknetChainId.SEPOLIA,
    )

    deploy_account_tx = await account.sign_deploy_account_v3(
        class_hash=class_hash,
        contract_address_salt=salt,
        constructor_calldata=[key_pair.public_key],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    resp = await account.client.deploy_account(transaction=deploy_account_tx)

    assert resp.address == address


@pytest.mark.asyncio
async def test_deploy_account_v3(client, deploy_account_details_factory):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    account = Account(
        address=address,
        client=client,
        key_pair=key_pair,
        chain=StarknetChainId.SEPOLIA,
    )

    deploy_account_tx = await account.sign_deploy_account_v3(
        class_hash=class_hash,
        contract_address_salt=salt,
        constructor_calldata=[key_pair.public_key],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    assert isinstance(deploy_account_tx, DeployAccountV3)

    resp = await account.client.deploy_account(transaction=deploy_account_tx)

    assert resp.address == address



import re
import sys
from enum import Enum
from typing import Any, Mapping, Optional, Union

from marshmallow import Schema, ValidationError, fields, post_load

from starknet_py.net.client_models import (
    BlockStatus,
    CallType,
    DAMode,
    EntryPointType,
    L1DAMode,
    PriceUnit,
    StorageEntry,
    TransactionExecutionStatus,
    TransactionFinalityStatus,
    TransactionStatus,
    TransactionStatusWithoutL1,
    TransactionType,
)

# pylint: disable=unused-argument


def _pascal_to_screaming_upper(checked_string: str) -> str:
    if bool(re.fullmatch(r"[A-Z0-9]+(?:_[A-Z0-9]+)*", checked_string)):
        return checked_string
    return re.sub(r"(?<!^)(?=[A-Z])", "_", checked_string).upper()


class NumberAsHex(fields.Field):
    """
    This field performs the following operations:

    - Serializes integers into hexadecimal strings
    - Deserializes hexadecimal strings into integers

    If a valid hexadecimal string is provided during serialization, it is returned as is.
    Similarly, when a valid integer is provided during deserialization, it remains unchanged.
    """

    MAX_VALUE = sys.maxsize
    REGEX_PATTERN = r"^0x[a-fA-F0-9]+$"

    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if self._is_int_and_in_range(value):
            return hex(value)

        if self._is_str_and_valid_pattern(value):
            return value

        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}"
        )

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Union[Mapping[str, Any], None],
        **kwargs,
    ):
        if self._is_int_and_in_range(value):
            return value

        if self._is_str_and_valid_pattern(value):
            return int(value, 16)

        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}"
        )

    def _is_int_and_in_range(self, value: Any) -> bool:
        return isinstance(value, int) and 0 <= value < self.MAX_VALUE

    def _is_str_and_valid_pattern(self, value: Any) -> bool:
        return (
            isinstance(value, str)
            and re.fullmatch(self.REGEX_PATTERN, value) is not None
        )


class Felt(NumberAsHex):
    """
    Field used to serialize and deserialize felt type.
    """

    MAX_VALUE = 2**252
    REGEX_PATTERN = r"^0x(0|[a-fA-F1-9]{1}[a-fA-F0-9]{0,62})$"


class Uint64(NumberAsHex):
    """
    Field used to serialize and deserialize RPC u64 type.
    """

    MAX_VALUE = 2**64
    REGEX_PATTERN = r"^0x(0|[a-fA-F1-9]{1}[a-fA-F0-9]{0,15})$"


class Uint128(NumberAsHex):
    """
    Field used to serialize and deserialize RPC u128 type.
    """

    MAX_VALUE = 2**128
    REGEX_PATTERN = r"^0x(0|[a-fA-F1-9]{1}[a-fA-F0-9]{0,31})$"


class NonPrefixedHex(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return hex(value).lstrip("0x")

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        return int(value, 16)


class StatusField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionStatus:
        values = [v.value for v in TransactionStatus]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionStatus: {value}."
            )

        return TransactionStatus(value)


class ExecutionStatusField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionExecutionStatus:
        values = [v.value for v in TransactionExecutionStatus]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionExecutionStatus: {value}."
            )

        return TransactionExecutionStatus(value)


class FinalityStatusField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionFinalityStatus:
        values = [v.value for v in TransactionFinalityStatus]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionFinalityStatus: {value}."
            )

        return TransactionFinalityStatus(value)


class TransactionFinalityStatusField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionFinalityStatus:
        values = [v.value for v in TransactionFinalityStatus]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionFinalityStatusField: {value}."
            )

        return TransactionFinalityStatus(value)


class TransactionStatusWithoutL1Field(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionStatusWithoutL1:
        values = [v.value for v in TransactionStatusWithoutL1]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionStatusWithoutL1: {value}."
            )

        return TransactionStatusWithoutL1(value)


class BlockStatusField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> BlockStatus:
        values = [v.value for v in BlockStatus]

        if value not in values:
            raise ValidationError(f"Invalid value for BlockStatus provided: {value}.")

        return BlockStatus(value)


class TransactionTypeField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if value == TransactionType.INVOKE:
            return "INVOKE_FUNCTION"
        return value.name

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> TransactionType:
        values = [v.value for v in TransactionType]

        if value == "INVOKE_FUNCTION":
            value = "INVOKE"

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for TransactionType: {value}."
            )

        return TransactionType(value)


class EntryPointTypeField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> EntryPointType:
        values = [v.value for v in EntryPointType]

        if value not in values:
            raise ValidationError(
                f"Invalid value provided for EntryPointType: {value}."
            )

        return EntryPointType(value)


class CallTypeField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> CallType:
        values = [v.value for v in CallType]

        if value not in values:
            raise ValidationError(f"Invalid value provided for CallType: {value}.")

        return CallType(value)


class L1DAModeField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> L1DAMode:
        values = [v.value for v in L1DAMode]

        if value not in values:
            raise ValidationError(f"Invalid value provided for L1DAMode: {value}.")

        return L1DAMode(value)


class PriceUnitField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> PriceUnit:
        values = [v.value for v in PriceUnit]

        if value not in values:
            raise ValidationError(f"Invalid value provided for PriceUnit: {value}.")

        return PriceUnit(value)


class DAModeField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        return value.name if value is not None else ""

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> DAMode:
        names = [v.name for v in DAMode]

        if value not in names:
            raise ValidationError(f"Invalid value provided for DAMode: {value}.")

        return DAMode[value]


class StorageEntrySchema(Schema):
    key = Felt(data_key="key", required=True)
    value = Felt(data_key="value", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        # pylint: disable=no-self-use
        return StorageEntry(**data)


class Revision(Enum):
    """
    Enum representing the revision of the specification to be used.
    """

    V0 = 0
    V1 = 1


class RevisionField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if value is None or value == Revision.V0:
            return str(Revision.V0.value)
        return value.value

    def _deserialize(self, value, attr, data, **kwargs) -> Revision:
        if isinstance(value, str):
            value = int(value)

        if isinstance(value, Revision):
            value = value.value

        revisions = [revision.value for revision in Revision]
        if value not in revisions:
            allowed_revisions_str = "".join(list(map(str, revisions)))
            raise ValidationError(
                f"Invalid value provided for Revision: {value}. Allowed values are {allowed_revisions_str}."
            )

        return Revision(value)



from marshmallow import fields, post_load

from starknet_py.devnet_utils.devnet_client_models import (
    Balance,
    BalanceRecord,
    Config,
    ForkConfig,
    IncreaseTimeResponse,
    MessageToL1,
    MessageToL2,
    MintResponse,
    PostmanFlushResponse,
    PredeployedAccount,
    ServerConfig,
    SetTimeResponse,
)
from starknet_py.net.schemas.common import Felt, PriceUnitField
from starknet_py.utils.schema import Schema

# pylint: disable=unused-argument, no-self-use


class MintResponseSchema(Schema):
    new_balance = fields.Integer(data_key="new_balance", required=True)
    unit = PriceUnitField(data_key="unit", required=True)
    tx_hash = Felt(data_key="tx_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> MintResponse:
        return MintResponse(**data)


class BalanceRecordSchema(Schema):
    amount = fields.Integer(data_key="amount", required=True)
    unit = PriceUnitField(data_key="unit", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> BalanceRecord:
        return BalanceRecord(**data)


class BalanceSchema(Schema):
    eth = fields.Nested(BalanceRecordSchema(), data_key="eth", required=True)
    strk = fields.Nested(BalanceRecordSchema(), data_key="strk", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Balance:
        return Balance(**data)


class MessageToL1Schema(Schema):
    from_address = Felt(data_key="from_address", required=True)
    to_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> MessageToL1:
        return MessageToL1(**data)


class MessageToL2Schema(Schema):
    l2_contract_address = Felt(data_key="l2_contract_address", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)
    l1_contract_address = Felt(data_key="l1_contract_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)
    paid_fee_on_l1 = Felt(data_key="paid_fee_on_l1", required=True)
    nonce = Felt(data_key="nonce", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> MessageToL2:
        return MessageToL2(**data)


class PostmanFlushResponseSchema(Schema):
    messages_to_l1 = fields.List(
        fields.Nested(MessageToL1Schema(), data_key="messages_to_l1", required=False)
    )
    messages_to_l2 = fields.List(
        fields.Nested(MessageToL2Schema(), data_key="messages_to_l2", required=False)
    )
    generated_l2_transactions = fields.List(
        Felt(), data_key="generated_l2_transactions", required=False
    )
    l1_provider = fields.String(data_key="l1_provider", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> PostmanFlushResponse:
        return PostmanFlushResponse(**data)


class PredeployedAccountSchema(Schema):
    initial_balance = fields.Integer(data_key="initial_balance", required=True)
    address = Felt(data_key="address", required=True)
    public_key = Felt(data_key="public_key", required=True)
    private_key = Felt(data_key="private_key", required=True)
    balance = fields.Nested(
        BalanceSchema(), data_key="balance", load_default=None, required=False
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> PredeployedAccount:
        return PredeployedAccount(**data)


class ForkConfigSchema(Schema):
    url = fields.String(data_key="url", required=False, load_default=None)
    block_number = fields.Integer(
        data_key="block_number", required=False, load_default=None
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ForkConfig:
        return ForkConfig(**data)


class ServerConfigSchema(Schema):
    host = fields.String(data_key="host", required=True)
    port = fields.Integer(data_key="port", required=True)
    timeout = fields.Integer(data_key="timeout", required=True)
    request_body_size_limit = fields.Integer(
        data_key="request_body_size_limit", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ServerConfig:
        return ServerConfig(**data)


class ConfigSchema(Schema):
    seed = fields.Integer(data_key="seed", required=True)
    total_accounts = fields.Integer(data_key="total_accounts", required=True)
    account_contract_class = fields.String(
        data_key="account_contract_class", required=False
    )
    account_contract_class_hash = Felt(
        data_key="account_contract_class_hash", required=False
    )
    predeployed_accounts_initial_balance = fields.String(
        data_key="predeployed_accounts_initial_balance", required=True
    )
    start_time = fields.Integer(
        data_key="start_time", required=False, load_default=None
    )
    gas_price_wei = fields.Integer(data_key="gas_price_wei", required=True)
    gas_price_strk = fields.Integer(data_key="gas_price_strk", required=True)
    data_gas_price_wei = fields.Integer(data_key="data_gas_price_wei", required=True)
    data_gas_price_strk = fields.Integer(data_key="data_gas_price_strk", required=True)
    chain_id = fields.String(data_key="chain_id", required=True)
    dump_on = fields.Integer(data_key="dump_on", required=False, load_default=None)
    dump_path = fields.String(data_key="dump_path", required=False, load_default=None)
    block_generation_on = fields.String(data_key="block_generation_on", required=True)
    lite_mode = fields.Boolean(data_key="lite_mode", required=True)
    state_archive = fields.String(data_key="state_archive", required=True)
    fork_config = fields.Nested(
        ForkConfigSchema(), data_key="fork_config", required=True
    )
    disable_account_impersonation = fields.Boolean(
        data_key="disable_account_impersonation", required=True
    )
    server_config = fields.Nested(
        ServerConfigSchema(), data_key="server_config", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Config:
        return Config(**data)


class IncreasedTimeResponseSchema(Schema):
    timestamp_increased_by = fields.Integer(
        data_key="timestamp_increased_by", required=True
    )
    block_hash = Felt(data_key="block_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> IncreaseTimeResponse:
        return IncreaseTimeResponse(**data)


class SetTimeResponseSchema(Schema):
    block_timestamp = fields.Integer(data_key="block_timestamp", required=True)
    block_hash = Felt(data_key="block_hash", required=False, load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SetTimeResponse:
        return SetTimeResponse(**data)



# pylint: disable=unused-variable
import pytest

from starknet_py.contract import Contract
from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping


@pytest.mark.asyncio
async def test_invoke(map_contract: Contract):
    prepared_function_call = map_contract.functions["put"].prepare_invoke_v3(
        key=10, value=20
    )
    # docs-start: invoke
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e20)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    invoke_result = await prepared_function_call.invoke(resource_bounds=resource_bounds)
    # docs-end: invoke
    prepared_function_call.resource_bounds = None
    # docs-start: invoke
    # resource_bounds can be estimated automatically
    invoke_result = await prepared_function_call.invoke(auto_estimate=True)
    # or if resource_bounds was specified in prepared_function_call
    # docs-end: invoke
    prepared_function_call.resource_bounds = resource_bounds
    # docs-start: invoke
    invoke_result = await prepared_function_call.invoke()
    # docs-end: invoke



from typing import List

import pytest


@pytest.mark.asyncio
async def test_custom_signer():
    # pylint: disable=import-outside-toplevel, duplicate-code, unused-variable

    # docs: start
    from starknet_py.net.account.account import Account
    from starknet_py.net.full_node_client import FullNodeClient
    from starknet_py.net.models import StarknetChainId, Transaction
    from starknet_py.net.signer import BaseSigner
    from starknet_py.utils.typed_data import TypedData

    # Create a custom signer class implementing BaseSigner interface
    class CustomSigner(BaseSigner):
        @property
        def public_key(self) -> int:
            return 0x123

        def sign_transaction(self, transaction: Transaction) -> List[int]:
            return [0x0, 0x1]

        def sign_message(
            self, typed_data: TypedData, account_address: int
        ) -> List[int]:
            return [0x0, 0x1]

    # Create an Account instance with the signer you've implemented
    custom_signer = CustomSigner()
    client = FullNodeClient(node_url="https://your.node.url")
    account = Account(
        client=client,
        address=0x1111,
        signer=custom_signer,
        chain=StarknetChainId.SEPOLIA,
    )
    # Now you can use Account as you'd always do
    # docs: end



import json

import pytest

import starknet_py.tests.e2e.fixtures.abi_structures as fixtures
from starknet_py.abi.v0 import AbiParser, AbiParsingError
from starknet_py.cairo.type_parser import UnknownCairoTypeError
from starknet_py.tests.e2e.fixtures.constants import CAIRO_0_CONTRACTS_ABI_DIR
from starknet_py.tests.e2e.fixtures.misc import read_contract


def test_parsing_types_abi():
    # Even though user depend on pool id and uint256 it is defined first. Parser has to consider those cases
    abi = AbiParser(
        [
            fixtures.user_dict,
            fixtures.pool_id_dict,
            fixtures.uint256_dict,
            fixtures.user_added_dict,
            fixtures.pool_id_added_dict,
            fixtures.get_user_dict,
            fixtures.delete_pool_dict,
        ]
    ).parse()

    assert abi.defined_structures == {
        "Uint256": fixtures.uint256_struct,
        "PoolId": fixtures.pool_id_struct,
        "User": fixtures.user_struct,
    }
    assert abi.events == {
        "UserAdded": fixtures.user_added_event,
        "PoolIdAdded": fixtures.pool_id_added_event,
    }
    assert abi.functions == {
        "get_user": fixtures.get_user_fn,
        "delete_pool": fixtures.delete_pool_fn,
    }


def test_parsing_types_abi_missing_offset():
    abi = AbiParser(
        [
            fixtures.user_missing_offset_dict,
            fixtures.uint256_dict,
            fixtures.pool_id_dict,
        ]
    ).parse()

    assert abi.defined_structures == {
        "Uint256": fixtures.uint256_struct,
        "PoolId": fixtures.pool_id_struct,
        "User": fixtures.user_struct,
    }


def test_parsing_types_abi_partial_missing_offset():
    abi = AbiParser(
        [
            fixtures.user_partial_missing_offset_dict,
            fixtures.uint256_dict,
            fixtures.pool_id_dict,
        ]
    ).parse()

    assert abi.defined_structures == {
        "Uint256": fixtures.uint256_struct,
        "PoolId": fixtures.pool_id_struct,
        "User": fixtures.user_partial_missing_offset_struct,
    }


def test_self_cycle():
    self_referencing_struct = {
        "type": "struct",
        "name": "Infinite",
        "size": 1,
        "members": [
            {"name": "value", "offset": 0, "type": "Infinite"},
        ],
    }
    with pytest.raises(
        AbiParsingError,
        match="Circular reference detected",
    ):
        AbiParser([self_referencing_struct]).parse()


def test_bigger_cycle():
    # first -> seconds -> third -> first...
    first = {
        "type": "struct",
        "name": "First",
        "size": 1,
        "members": [{"name": "value", "offset": 0, "type": "Second"}],
    }
    second = {
        "type": "struct",
        "name": "Second",
        "size": 1,
        "members": [{"name": "value", "offset": 0, "type": "Third"}],
    }
    third = {
        "type": "struct",
        "name": "Third",
        "size": 1,
        "members": [{"name": "value", "offset": 0, "type": "First"}],
    }
    with pytest.raises(
        AbiParsingError,
        match="Circular reference detected",
    ):
        AbiParser([first, second, third]).parse()


def test_duplicated_structure():
    with pytest.raises(
        AbiParsingError,
        match="Name 'Uint256' was used more than once in defined structures",
    ):
        AbiParser(
            [fixtures.uint256_dict, fixtures.pool_id_dict, fixtures.uint256_dict]
        ).parse()


def test_duplicated_function():
    with pytest.raises(
        AbiParsingError,
        match="Name 'get_user' was used more than once in defined functions",
    ):
        AbiParser(
            [
                fixtures.get_user_dict,
                fixtures.delete_pool_dict,
                fixtures.get_user_dict,
                fixtures.delete_pool_dict,
            ]
        ).parse()


def test_duplicated_event():
    with pytest.raises(
        AbiParsingError,
        match="Name 'UserAdded' was used more than once in defined events",
    ):
        AbiParser(
            [
                fixtures.user_added_dict,
                fixtures.delete_pool_dict,
                fixtures.user_added_dict,
            ]
        ).parse()


def test_duplicated_type_members():
    type_dict = {
        "type": "struct",
        "name": "Record",
        "size": 4,
        "members": [
            {"name": "name", "offset": 0, "type": "felt"},
            {"name": "value", "offset": 1, "type": "felt"},
            {"name": "id", "offset": 2, "type": "felt"},
            {"name": "value", "offset": 3, "type": "felt"},
        ],
    }
    with pytest.raises(
        AbiParsingError,
        match="Name 'value' was used more than once in members of structure 'Record'",
    ):
        AbiParser([type_dict]).parse()


@pytest.mark.parametrize(
    "missing_name, input_dict",
    [
        # Type
        ("Uint256", fixtures.pool_id_dict),
        # Function
        ("Uint256", fixtures.get_user_dict),
        # Event
        ("User", fixtures.user_added_dict),
    ],
)
def test_missing_type_used(missing_name, input_dict):
    with pytest.raises(
        UnknownCairoTypeError, match=f"Type '{missing_name}' is not defined"
    ):
        AbiParser([input_dict]).parse()


def test_deserialize_balance_struct_event_abi():
    # Contains all types of ABI apart from structures
    abi = json.loads(
        read_contract(
            "balance_struct_event_abi.json", directory=CAIRO_0_CONTRACTS_ABI_DIR
        )
    )

    deserialized = AbiParser(abi).parse()

    assert deserialized == fixtures.balance_struct_abi


def test_duplicated_constructor():
    constructor = {
        "inputs": [],
        "name": "constructor",
        "outputs": [],
        "type": "constructor",
    }
    with pytest.raises(
        AbiParsingError, match="Constructor in ABI must be defined at most once"
    ):
        AbiParser([constructor, constructor]).parse()


def test_duplicated_l1_handler():
    l1_handler = {
        "inputs": [],
        "name": "__l1_default__",
        "outputs": [],
        "type": "l1_handler",
    }
    with pytest.raises(
        AbiParsingError, match="L1 handler in ABI must be defined at most once"
    ):
        AbiParser([l1_handler, l1_handler]).parse()



import sys

import pytest

from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract


# TODO (#1219): investigate why this test fails for v1 contract
@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Some cairo 1 contracts compiled with v1 compiler fail with new devnet-rs - test simply for a code example.",
)
@pytest.mark.asyncio
async def test_simple_deploy_cairo1(account, erc20_class_hash):
    # pylint: disable=import-outside-toplevel
    # docs: start
    from starknet_py.cairo.felt import encode_shortstring
    from starknet_py.common import create_sierra_compiled_contract
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

    # docs: end

    compiled_contract = load_contract(
        contract_name="ERC20", version=ContractVersion.V2
    )["sierra"]

    class_hash = erc20_class_hash

    # docs: start
    abi = create_sierra_compiled_contract(
        compiled_contract=compiled_contract
    ).parsed_abi

    constructor_args = {
        "name_": encode_shortstring("erc20_basic"),
        "symbol_": encode_shortstring("ERC20B"),
        "decimals_": 10,
        "initial_supply": 12345,
        "recipient": account.address,
    }

    deploy_result = await Contract.deploy_contract_v3(
        account=account,
        class_hash=class_hash,
        abi=abi,
        constructor_args=constructor_args,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )

    await deploy_result.wait_for_acceptance()
    contract = deploy_result.deployed_contract
    # docs: end

    assert contract.address != 0



from typing import List, Optional, Union

import starknet_py.abi.v2.shape as ShapeV2
from starknet_py.abi.v0 import AbiParser as AbiParserV0
from starknet_py.abi.v1 import AbiParser as AbiParserV1
from starknet_py.abi.v2 import AbiParser as AbiParserV2
from starknet_py.serialization import (
    FunctionSerializationAdapter,
    serializer_for_function,
)
from starknet_py.serialization.factory import (
    serializer_for_constructor_v2,
    serializer_for_function_v1,
)


def translate_constructor_args(
    abi: List, constructor_args: Optional[Union[List, dict]], *, cairo_version: int = 1
) -> List[int]:
    serializer = (
        _get_constructor_serializer_v1(abi)
        if cairo_version == 1
        else _get_constructor_serializer_v0(abi)
    )

    if serializer is None or len(serializer.inputs_serializer.serializers) == 0:
        return []

    if not constructor_args:
        raise ValueError(
            "Provided contract has a constructor and no arguments were provided."
        )

    args, kwargs = (
        ([], constructor_args)
        if isinstance(constructor_args, dict)
        else (constructor_args, {})
    )
    return serializer.serialize(*args, **kwargs)


def _get_constructor_serializer_v1(abi: List) -> Optional[FunctionSerializationAdapter]:
    if _is_abi_v2(abi):
        parsed = AbiParserV2(abi).parse()
        constructor = parsed.constructor

        if constructor is None or not constructor.inputs:
            return None

        return serializer_for_constructor_v2(constructor)

    parsed = AbiParserV1(abi).parse()
    constructor = parsed.functions.get("constructor", None)

    # Constructor might not accept any arguments
    if constructor is None or not constructor.inputs:
        return None

    return serializer_for_function_v1(constructor)


def _is_abi_v2(abi: List) -> bool:
    for entry in abi:
        if entry["type"] in [
            ShapeV2.CONSTRUCTOR_ENTRY,
            ShapeV2.L1_HANDLER_ENTRY,
            ShapeV2.INTERFACE_ENTRY,
            ShapeV2.IMPL_ENTRY,
        ]:
            return True
        if entry["type"] == ShapeV2.EVENT_ENTRY:
            if "inputs" in entry:
                return False
            if "kind" in entry:
                return True
    return False


def _get_constructor_serializer_v0(abi: List) -> Optional[FunctionSerializationAdapter]:
    parsed = AbiParserV0(abi).parse()

    # Constructor might not accept any arguments
    if not parsed.constructor or not parsed.constructor.inputs:
        return None

    return serializer_for_function(parsed.constructor)



from typing import Any, Optional

from marshmallow import ValidationError, fields, post_load, validate

from starknet_py.net.executable_models import (
    AllocConstantSize,
    AllocConstantSizeInner,
    AllocFelt252Dict,
    AllocFelt252DictInner,
    AllocSegment,
    AllocSegmentInner,
    AssertAllAccessesUsed,
    AssertAllAccessesUsedInner,
    AssertLeFindSmallArcs,
    AssertLeFindSmallArcsInner,
    AssertLeIsFirstArcExcluded,
    AssertLeIsFirstArcExcludedInner,
    AssertLeIsSecondArcExcluded,
    AssertLeIsSecondArcExcludedInner,
    AssertLtAssertValidInput,
    AssertLtAssertValidInputInner,
    BinOp,
    BinOpInner,
    CasmClass,
    CasmClassEntryPointsByType,
    CellRef,
    Cheatcode,
    CheatcodeInner,
    DebugPrint,
    DebugPrintInner,
    Deref,
    DivMod,
    DivModInner,
    DoubleDeref,
    EvalCircuit,
    EvalCircuitInner,
    Felt252DictEntryInit,
    Felt252DictEntryInitInner,
    Felt252DictEntryUpdate,
    Felt252DictEntryUpdateInner,
    Felt252DictRead,
    Felt252DictReadInner,
    Felt252DictWrite,
    Felt252DictWriteInner,
    FieldSqrt,
    FieldSqrtInner,
    GetCurrentAccessDelta,
    GetCurrentAccessDeltaInner,
    GetCurrentAccessIndex,
    GetCurrentAccessIndexInner,
    GetNextDictKey,
    GetNextDictKeyInner,
    GetSegmentArenaIndex,
    GetSegmentArenaIndexInner,
    Hint,
    Immediate,
    InitSquashData,
    InitSquashDataInner,
    LinearSplit,
    LinearSplitInner,
    RandomEcPoint,
    RandomEcPointInner,
    ShouldContinueSquashLoop,
    ShouldContinueSquashLoopInner,
    ShouldSkipSquashLoop,
    ShouldSkipSquashLoopInner,
    SquareRoot,
    SquareRootInner,
    SystemCall,
    SystemCallInner,
    TestLessThan,
    TestLessThanInner,
    TestLessThanOrEqual,
    TestLessThanOrEqualAddress,
    TestLessThanOrEqualAddressInner,
    TestLessThanOrEqualInner,
    U256InvModN,
    U256InvModNInner,
    Uint256DivMod,
    Uint256DivModInner,
    Uint256SquareRoot,
    Uint256SquareRootInner,
    Uint512DivModByUint256,
    Uint512DivModByUint256Inner,
    WideMul128,
    WideMul128Inner,
)
from starknet_py.net.schemas.common import NumberAsHex
from starknet_py.utils.schema import Schema


class CellRefSchema(Schema):
    register = fields.String(
        data_key="register", validate=validate.OneOf(["AP", "FP"]), required=True
    )
    offset = fields.Integer(data_key="offset", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> CellRef:
        return CellRef(**data)


class AssertAllAccessesUsedInnerSchema(Schema):
    n_used_accesses = fields.Nested(
        CellRefSchema(), data_key="n_used_accesses", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertAllAccessesUsedInner:
        return AssertAllAccessesUsedInner(**data)


class AssertAllAccessesUsedSchema(Schema):
    assert_all_accesses_used = fields.Nested(
        AssertAllAccessesUsedInnerSchema(),
        data_key="AssertAllAccessesUsed",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertAllAccessesUsed:
        return AssertAllAccessesUsed(**data)


class DerefSchema(Schema):
    deref = fields.Nested(CellRefSchema(), data_key="Deref", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Deref:
        return Deref(**data)


class DoubleDerefSchema(Schema):
    double_deref = fields.Tuple(
        (fields.Nested(CellRefSchema()), fields.Integer()),
        data_key="DoubleDeref",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DoubleDeref:
        return DoubleDeref(**data)


class ImmediateSchema(Schema):
    immediate = NumberAsHex(data_key="Immediate", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Immediate:
        return Immediate(**data)


class BinOpBField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if isinstance(value, Deref):
            return DerefSchema().dump(value)
        elif isinstance(value, Immediate):
            return ImmediateSchema().dump(value)

        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}."
        )

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dict):
            if "Deref" in value:
                return DerefSchema().load(value)
            elif "Immediate" in value:
                return ImmediateSchema().load(value)

        raise ValidationError(
            f"Invalid value provided for 'b': {value}. Must be a Deref or an Immediate object."
        )


class BinOpInnerSchema(Schema):
    op = fields.String(
        data_key="op", required=True, validate=validate.OneOf(["Add", "Mul"])
    )
    a = fields.Nested(CellRefSchema(), data_key="a", required=True)
    b = BinOpBField(data_key="b", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> BinOpInner:
        return BinOpInner(**data)


class BinOpSchema(Schema):
    bin_op = fields.Nested(BinOpInnerSchema(), data_key="BinOp", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> BinOp:
        return BinOp(**data)


class ResOperandField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, Deref):
            return DerefSchema().dump(value)
        elif isinstance(value, DoubleDeref):
            return DoubleDerefSchema().dump(value)
        elif isinstance(value, Immediate):
            return ImmediateSchema().dump(value)
        elif isinstance(value, BinOp):
            return BinOpSchema().dump(value)

        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}."
        )

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dict):
            if "Deref" in value:
                return DerefSchema().load(value)
            elif "DoubleDeref" in value:
                return DoubleDerefSchema().load(value)
            elif "Immediate" in value:
                return ImmediateSchema().load(value)
            elif "BinOp" in value:
                return BinOpSchema().load(value)

        raise ValidationError(f"Invalid value provided for ResOperand: {value}.")


class AssertLtAssertValidInputInnerSchema(Schema):
    a = ResOperandField(data_key="a", required=True)
    b = ResOperandField(data_key="b", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLtAssertValidInputInner:
        return AssertLtAssertValidInputInner(**data)


class AssertLtAssertValidInputSchema(Schema):
    assert_lt_assert_valid_input = fields.Nested(
        AssertLtAssertValidInputInnerSchema(),
        data_key="AssertLtAssertValidInput",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLtAssertValidInput:
        return AssertLtAssertValidInput(**data)


class Felt252DictReadInnerSchema(Schema):
    dict_ptr = ResOperandField(data_key="dict_ptr", required=True)
    key = ResOperandField(data_key="key", required=True)
    value_dst = fields.Nested(CellRefSchema(), data_key="value_dst", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictReadInner:
        return Felt252DictReadInner(**data)


class Felt252DictReadSchema(Schema):
    felt252_dict_read = fields.Nested(
        Felt252DictReadInnerSchema(), data_key="Felt252DictRead", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictRead:
        return Felt252DictRead(**data)


class Felt252DictWriteInnerSchema(Schema):
    dict_ptr = ResOperandField(data_key="dict_ptr", required=True)
    key = ResOperandField(data_key="key", required=True)
    value = ResOperandField(data_key="value", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictWriteInner:
        return Felt252DictWriteInner(**data)


class Felt252DictWriteSchema(Schema):
    felt252_dict_write = fields.Nested(
        Felt252DictWriteInnerSchema(), data_key="Felt252DictWrite", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictWrite:
        return Felt252DictWrite(**data)


class AllocSegmentInnerSchema(Schema):
    dst = fields.Nested(CellRefSchema(), data_key="dst", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocSegmentInner:
        return AllocSegmentInner(**data)


class AllocSegmentSchema(Schema):
    alloc_segment = fields.Nested(
        AllocSegmentInnerSchema(), data_key="AllocSegment", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocSegment:
        return AllocSegment(**data)


class TestLessThanInnerSchema(Schema):
    lhs = ResOperandField(data_key="lhs", required=True)
    rhs = ResOperandField(data_key="rhs", required=True)
    dst = fields.Nested(CellRefSchema(), data_key="dst", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThanInner:
        return TestLessThanInner(**data)


class TestLessThanSchema(Schema):
    test_less_than = fields.Nested(
        TestLessThanInnerSchema(), data_key="TestLessThan", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThan:
        return TestLessThan(**data)


class TestLessThanOrEqualInnerSchema(TestLessThanInnerSchema):
    pass

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThanOrEqualInner:
        return TestLessThanOrEqualInner(**data)


class TestLessThanOrEqualSchema(Schema):
    test_less_than_or_equal = fields.Nested(
        TestLessThanOrEqualInnerSchema(), data_key="TestLessThanOrEqual", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThanOrEqual:
        return TestLessThanOrEqual(**data)


class TestLessThanOrEqualAddressInnerSchema(TestLessThanInnerSchema):
    pass

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThanOrEqualAddressInner:
        return TestLessThanOrEqualAddressInner(**data)


class TestLessThanOrEqualAddressSchema(Schema):
    test_less_than_or_equal_address = fields.Nested(
        TestLessThanOrEqualAddressInnerSchema(),
        data_key="TestLessThanOrEqualAddress",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TestLessThanOrEqualAddress:
        return TestLessThanOrEqualAddress(**data)


class WideMul128InnerSchema(Schema):
    lhs = ResOperandField(data_key="lhs", required=True)
    rhs = ResOperandField(data_key="rhs", required=True)
    high = fields.Nested(CellRefSchema(), data_key="high", required=True)
    low = fields.Nested(CellRefSchema(), data_key="low", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> WideMul128Inner:
        return WideMul128Inner(**data)


class WideMul128Schema(Schema):
    wide_mul128 = fields.Nested(
        WideMul128InnerSchema(), data_key="WideMul128", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> WideMul128:
        return WideMul128(**data)


class DivModInnerSchema(Schema):
    lhs = ResOperandField(data_key="lhs", required=True)
    rhs = ResOperandField(data_key="rhs", required=True)
    quotient = fields.Nested(CellRefSchema(), data_key="quotient", required=True)
    remainder = fields.Nested(CellRefSchema(), data_key="remainder", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DivModInner:
        return DivModInner(**data)


class DivModSchema(Schema):
    div_mod = fields.Nested(DivModInnerSchema(), data_key="DivMod", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DivMod:
        return DivMod(**data)


class Uint256DivModInnerSchema(Schema):
    dividend0 = ResOperandField(data_key="dividend0", required=True)
    dividend1 = ResOperandField(data_key="dividend1", required=True)
    divisor0 = ResOperandField(data_key="divisor0", required=True)
    divisor1 = ResOperandField(data_key="divisor1", required=True)
    quotient0 = fields.Nested(CellRefSchema(), data_key="quotient0", required=True)
    quotient1 = fields.Nested(CellRefSchema(), data_key="quotient1", required=True)
    remainder0 = fields.Nested(CellRefSchema(), data_key="remainder0", required=True)
    remainder1 = fields.Nested(CellRefSchema(), data_key="remainder1", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint256DivModInner:
        return Uint256DivModInner(**data)


class Uint256DivModSchema(Schema):
    uint256_div_mod = fields.Nested(
        Uint256DivModInnerSchema(), data_key="Uint256DivMod", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint256DivMod:
        return Uint256DivMod(**data)


class Uint512DivModByUint256InnerSchema(Schema):
    dividend0 = ResOperandField(data_key="dividend0", required=True)
    dividend1 = ResOperandField(data_key="dividend1", required=True)
    dividend2 = ResOperandField(data_key="dividend2", required=True)
    dividend3 = ResOperandField(data_key="dividend3", required=True)
    divisor0 = ResOperandField(data_key="divisor0", required=True)
    divisor1 = ResOperandField(data_key="divisor1", required=True)
    quotient0 = fields.Nested(CellRefSchema(), data_key="quotient0", required=True)
    quotient1 = fields.Nested(CellRefSchema(), data_key="quotient1", required=True)
    quotient2 = fields.Nested(CellRefSchema(), data_key="quotient2", required=True)
    quotient3 = fields.Nested(CellRefSchema(), data_key="quotient3", required=True)
    remainder0 = fields.Nested(CellRefSchema(), data_key="remainder0", required=True)
    remainder1 = fields.Nested(CellRefSchema(), data_key="remainder1", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint512DivModByUint256Inner:
        return Uint512DivModByUint256Inner(**data)


class Uint512DivModByUint256Schema(Schema):
    uint512_div_mod_by_uint256 = fields.Nested(
        Uint512DivModByUint256InnerSchema(),
        data_key="Uint512DivModByUint256",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint512DivModByUint256:
        return Uint512DivModByUint256(**data)


class SquareRootInnerSchema(Schema):
    value = ResOperandField(data_key="value", required=True)
    dst = fields.Nested(CellRefSchema(), data_key="dst", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SquareRootInner:
        return SquareRootInner(**data)


class SquareRootSchema(Schema):
    square_root = fields.Nested(
        SquareRootInnerSchema(), data_key="SquareRoot", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> SquareRoot:
        return SquareRoot(**data)


class Uint256SquareRootInnerSchema(Schema):
    value_low = ResOperandField(data_key="value_low", required=True)
    value_high = ResOperandField(data_key="value_high", required=True)
    sqrt0 = fields.Nested(CellRefSchema(), data_key="sqrt0", required=True)
    sqrt1 = fields.Nested(CellRefSchema(), data_key="sqrt1", required=True)
    remainder_low = fields.Nested(
        CellRefSchema(), data_key="remainder_low", required=True
    )
    remainder_high = fields.Nested(
        CellRefSchema(), data_key="remainder_high", required=True
    )
    sqrt_mul_2_minus_remainder_ge_u128 = fields.Nested(
        CellRefSchema(), data_key="sqrt_mul_2_minus_remainder_ge_u128", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint256SquareRootInner:
        return Uint256SquareRootInner(**data)


class Uint256SquareRootSchema(Schema):
    uint256_square_root = fields.Nested(
        Uint256SquareRootInnerSchema(), data_key="Uint256SquareRoot", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Uint256SquareRoot:
        return Uint256SquareRoot(**data)


class LinearSplitInnerSchema(Schema):
    value = ResOperandField(data_key="value", required=True)
    scalar = ResOperandField(data_key="scalar", required=True)
    max_x = ResOperandField(data_key="max_x", required=True)
    x = fields.Nested(CellRefSchema(), data_key="x", required=True)
    y = fields.Nested(CellRefSchema(), data_key="y", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> LinearSplitInner:
        return LinearSplitInner(**data)


class LinearSplitSchema(Schema):
    linear_split = fields.Nested(
        LinearSplitInnerSchema(), data_key="LinearSplit", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> LinearSplit:
        return LinearSplit(**data)


class AllocFelt252DictInnerSchema(Schema):
    segment_arena_ptr = ResOperandField(data_key="segment_arena_ptr", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocFelt252DictInner:
        return AllocFelt252DictInner(**data)


class AllocFelt252DictSchema(Schema):
    alloc_felt252_dict = fields.Nested(
        AllocFelt252DictInnerSchema(), data_key="AllocFelt252Dict", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocFelt252Dict:
        return AllocFelt252Dict(**data)


class Felt252DictEntryInitInnerSchema(Schema):
    dict_ptr = ResOperandField(data_key="dict_ptr", required=True)
    key = ResOperandField(data_key="key", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictEntryInitInner:
        return Felt252DictEntryInitInner(**data)


class Felt252DictEntryInitSchema(Schema):
    felt252_dict_entry_init = fields.Nested(
        Felt252DictEntryInitInnerSchema(),
        data_key="Felt252DictEntryInit",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictEntryInit:
        return Felt252DictEntryInit(**data)


class Felt252DictEntryUpdateInnerSchema(Schema):
    dict_ptr = ResOperandField(data_key="dict_ptr", required=True)
    value = ResOperandField(data_key="value", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictEntryUpdateInner:
        return Felt252DictEntryUpdateInner(**data)


class Felt252DictEntryUpdateSchema(Schema):
    felt252_dict_entry_update = fields.Nested(
        Felt252DictEntryUpdateInnerSchema(),
        data_key="Felt252DictEntryUpdate",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Felt252DictEntryUpdate:
        return Felt252DictEntryUpdate(**data)


class GetSegmentArenaIndexInnerSchema(Schema):
    dict_end_ptr = ResOperandField(data_key="dict_end_ptr", required=True)
    dict_index = fields.Nested(CellRefSchema(), data_key="dict_index", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetSegmentArenaIndexInner:
        return GetSegmentArenaIndexInner(**data)


class GetSegmentArenaIndexSchema(Schema):
    get_segment_arena_index = fields.Nested(
        GetSegmentArenaIndexInnerSchema(),
        data_key="GetSegmentArenaIndex",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetSegmentArenaIndex:
        return GetSegmentArenaIndex(**data)


class InitSquashDataInnerSchema(Schema):
    dict_accesses = ResOperandField(data_key="dict_accesses", required=True)
    ptr_diff = ResOperandField(data_key="ptr_diff", required=True)
    n_accesses = ResOperandField(data_key="n_accesses", required=True)
    big_keys = fields.Nested(CellRefSchema(), data_key="big_keys", required=True)
    first_key = fields.Nested(CellRefSchema(), data_key="first_key", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> InitSquashDataInner:
        return InitSquashDataInner(**data)


class InitSquashDataSchema(Schema):
    init_squash_data = fields.Nested(
        InitSquashDataInnerSchema(), data_key="InitSquashData", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> InitSquashData:
        return InitSquashData(**data)


class GetCurrentAccessIndexInnerSchema(Schema):
    range_check_ptr = ResOperandField(data_key="range_check_ptr", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetCurrentAccessIndexInner:
        return GetCurrentAccessIndexInner(**data)


class GetCurrentAccessIndexSchema(Schema):
    get_current_access_index = fields.Nested(
        GetCurrentAccessIndexInnerSchema(),
        data_key="GetCurrentAccessIndex",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetCurrentAccessIndex:
        return GetCurrentAccessIndex(**data)


class ShouldSkipSquashLoopInnerSchema(Schema):
    should_skip_loop = fields.Nested(
        CellRefSchema(), data_key="should_skip_loop", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ShouldSkipSquashLoopInner:
        return ShouldSkipSquashLoopInner(**data)


class ShouldSkipSquashLoopSchema(Schema):
    should_skip_squash_loop = fields.Nested(
        ShouldSkipSquashLoopInnerSchema(),
        data_key="ShouldSkipSquashLoop",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ShouldSkipSquashLoop:
        return ShouldSkipSquashLoop(**data)


class GetCurrentAccessDeltaInnerSchema(Schema):
    index_delta_minus1 = fields.Nested(
        CellRefSchema(), data_key="index_delta_minus1", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetCurrentAccessDeltaInner:
        return GetCurrentAccessDeltaInner(**data)


class GetCurrentAccessDeltaSchema(Schema):
    get_current_access_delta = fields.Nested(
        GetCurrentAccessDeltaInnerSchema(),
        data_key="GetCurrentAccessDelta",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetCurrentAccessDelta:
        return GetCurrentAccessDelta(**data)


class ShouldContinueSquashLoopInnerSchema(Schema):
    should_continue = fields.Nested(
        CellRefSchema(), data_key="should_continue", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ShouldContinueSquashLoopInner:
        return ShouldContinueSquashLoopInner(**data)


class ShouldContinueSquashLoopSchema(Schema):
    should_continue_squash_loop = fields.Nested(
        ShouldContinueSquashLoopInnerSchema(),
        data_key="ShouldContinueSquashLoop",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ShouldContinueSquashLoop:
        return ShouldContinueSquashLoop(**data)


class GetNextDictKeyInnerSchema(Schema):
    next_key = fields.Nested(CellRefSchema(), data_key="next_key", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetNextDictKeyInner:
        return GetNextDictKeyInner(**data)


class GetNextDictKeySchema(Schema):
    get_next_dict_key = fields.Nested(
        GetNextDictKeyInnerSchema(), data_key="GetNextDictKey", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> GetNextDictKey:
        return GetNextDictKey(**data)


class AssertLeFindSmallArcsInnerSchema(Schema):
    range_check_ptr = ResOperandField(data_key="range_check_ptr", required=True)
    a = ResOperandField(data_key="a", required=True)
    b = ResOperandField(data_key="b", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeFindSmallArcsInner:
        return AssertLeFindSmallArcsInner(**data)


class AssertLeFindSmallArcsSchema(Schema):
    assert_le_find_small_arcs = fields.Nested(
        AssertLeFindSmallArcsInnerSchema(),
        data_key="AssertLeFindSmallArcs",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeFindSmallArcs:
        return AssertLeFindSmallArcs(**data)


class AssertLeIsFirstArcExcludedInnerSchema(Schema):
    skip_exclude_a_flag = fields.Nested(
        CellRefSchema(), data_key="skip_exclude_a_flag", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeIsFirstArcExcludedInner:
        return AssertLeIsFirstArcExcludedInner(**data)


class AssertLeIsFirstArcExcludedSchema(Schema):
    assert_le_is_first_arc_excluded = fields.Nested(
        AssertLeIsFirstArcExcludedInnerSchema(),
        data_key="AssertLeIsFirstArcExcluded",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeIsFirstArcExcluded:
        return AssertLeIsFirstArcExcluded(**data)


class AssertLeIsSecondArcExcludedInnerSchema(Schema):
    skip_exclude_b_minus_a = fields.Nested(
        CellRefSchema(), data_key="skip_exclude_b_minus_a", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeIsSecondArcExcludedInner:
        return AssertLeIsSecondArcExcludedInner(**data)


class AssertLeIsSecondArcExcludedSchema(Schema):
    assert_le_is_second_arc_excluded = fields.Nested(
        AssertLeIsSecondArcExcludedInnerSchema(),
        data_key="AssertLeIsSecondArcExcluded",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AssertLeIsSecondArcExcluded:
        return AssertLeIsSecondArcExcluded(**data)


class RandomEcPointInnerSchema(Schema):
    x = fields.Nested(CellRefSchema(), data_key="x", required=True)
    y = fields.Nested(CellRefSchema(), data_key="y", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> RandomEcPointInner:
        return RandomEcPointInner(**data)


class RandomEcPointSchema(Schema):
    random_ec_point = fields.Nested(
        RandomEcPointInnerSchema(), data_key="RandomEcPoint", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> RandomEcPoint:
        return RandomEcPoint(**data)


class FieldSqrtInnerSchema(Schema):
    val = ResOperandField(data_key="val", required=True)
    sqrt = fields.Nested(CellRefSchema(), data_key="sqrt", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> FieldSqrtInner:
        return FieldSqrtInner(**data)


class FieldSqrtSchema(Schema):
    field_sqrt = fields.Nested(
        FieldSqrtInnerSchema(), data_key="FieldSqrt", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> FieldSqrt:
        return FieldSqrt(**data)


class DebugPrintInnerSchema(Schema):
    start = ResOperandField(data_key="start", required=True)
    end = ResOperandField(data_key="end", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DebugPrintInner:
        return DebugPrintInner(**data)


class DebugPrintSchema(Schema):
    debug_print = fields.Nested(
        DebugPrintInnerSchema(), data_key="DebugPrint", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DebugPrint:
        return DebugPrint(**data)


class AllocConstantSizeInnerSchema(Schema):
    size = ResOperandField(data_key="size", required=True)
    dst = fields.Nested(CellRefSchema(), data_key="dst", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocConstantSizeInner:
        return AllocConstantSizeInner(**data)


class AllocConstantSizeSchema(Schema):
    alloc_constant_size = fields.Nested(
        AllocConstantSizeInnerSchema(), data_key="AllocConstantSize", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> AllocConstantSize:
        return AllocConstantSize(**data)


class U256InvModNInnerSchema(Schema):
    b0 = ResOperandField(data_key="b0", required=True)
    b1 = ResOperandField(data_key="b1", required=True)
    n0 = ResOperandField(data_key="n0", required=True)
    n1 = ResOperandField(data_key="n1", required=True)
    g0_or_no_inv = fields.Nested(
        CellRefSchema(), data_key="g0_or_no_inv", required=True
    )
    g1_option = fields.Nested(CellRefSchema(), data_key="g1_option", required=True)
    s_or_r0 = fields.Nested(CellRefSchema(), data_key="s_or_r0", required=True)
    s_or_r1 = fields.Nested(CellRefSchema(), data_key="s_or_r1", required=True)
    t_or_k0 = fields.Nested(CellRefSchema(), data_key="t_or_k0", required=True)
    t_or_k1 = fields.Nested(CellRefSchema(), data_key="t_or_k1", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> U256InvModNInner:
        return U256InvModNInner(**data)


class U256InvModNSchema(Schema):
    u256_inv_mod_n = fields.Nested(
        U256InvModNInnerSchema(), data_key="U256InvModN", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> U256InvModN:
        return U256InvModN(**data)


class EvalCircuitInnerSchema(Schema):
    n_add_mods = ResOperandField(data_key="n_add_mods", required=True)
    add_mod_builtin = ResOperandField(data_key="add_mod_builtin", required=True)
    n_mul_mods = ResOperandField(data_key="n_mul_mods", required=True)
    mul_mod_builtin = ResOperandField(data_key="mul_mod_builtin", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EvalCircuitInner:
        return EvalCircuitInner(**data)


class EvalCircuitSchema(Schema):
    eval_circuit = fields.Nested(
        EvalCircuitInnerSchema(), data_key="EvalCircuit", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> EvalCircuit:
        return EvalCircuit(**data)


class SystemCallInnerSchema(Schema):
    system = ResOperandField(data_key="system", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SystemCallInner:
        return SystemCallInner(**data)


class SystemCallSchema(Schema):
    system_call = fields.Nested(
        SystemCallInnerSchema(), data_key="SystemCall", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> SystemCall:
        return SystemCall(**data)


class CheatcodeInnerSchema(Schema):
    selector = NumberAsHex(data_key="selector", required=True)
    input_start = ResOperandField(data_key="input_start", required=True)
    input_end = ResOperandField(data_key="input_end", required=True)
    output_start = fields.Nested(
        CellRefSchema(), data_key="output_start", required=True
    )
    output_end = fields.Nested(CellRefSchema(), data_key="output_end", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> CheatcodeInner:
        return CheatcodeInner(**data)


class CheatcodeSchema(Schema):
    cheatcode = fields.Nested(
        CheatcodeInnerSchema(), data_key="Cheatcode", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> Cheatcode:
        return Cheatcode(**data)


HINT_TYPE_SCHEMAS_MAPPING = {
    "AllocConstantSize": AllocConstantSizeSchema,
    "AllocFelt252Dict": AllocFelt252DictSchema,
    "AllocSegment": AllocSegmentSchema,
    "AssertAllAccessesUsed": AssertAllAccessesUsedSchema,
    "AssertLeFindSmallArcs": AssertLeFindSmallArcsSchema,
    "AssertLeIsFirstArcExcluded": AssertLeIsFirstArcExcludedSchema,
    "AssertLeIsSecondArcExcluded": AssertLeIsSecondArcExcludedSchema,
    "AssertLtAssertValidInput": AssertLtAssertValidInputSchema,
    "BinOp": BinOpSchema,
    "Cheatcode": CheatcodeSchema,
    "DebugPrint": DebugPrintSchema,
    "Deref": DerefSchema,
    "DivMod": DivModSchema,
    "DoubleDeref": DoubleDerefSchema,
    "EvalCircuit": EvalCircuitSchema,
    "Felt252DictEntryInit": Felt252DictEntryInitSchema,
    "Felt252DictEntryUpdate": Felt252DictEntryUpdateSchema,
    "Felt252DictRead": Felt252DictReadSchema,
    "Felt252DictWrite": Felt252DictWriteSchema,
    "FieldSqrt": FieldSqrtSchema,
    "GetCurrentAccessDelta": GetCurrentAccessDeltaSchema,
    "GetCurrentAccessIndex": GetCurrentAccessIndexSchema,
    "GetNextDictKey": GetNextDictKeySchema,
    "GetSegmentArenaIndex": GetSegmentArenaIndexSchema,
    "Immediate": ImmediateSchema,
    "InitSquashData": InitSquashDataSchema,
    "LinearSplit": LinearSplitSchema,
    "RandomEcPoint": RandomEcPointSchema,
    "ShouldContinueSquashLoop": ShouldContinueSquashLoopSchema,
    "ShouldSkipSquashLoop": ShouldSkipSquashLoopSchema,
    "SquareRoot": SquareRootSchema,
    "SystemCall": SystemCallSchema,
    "TestLessThan": TestLessThanSchema,
    "TestLessThanOrEqual": TestLessThanOrEqualSchema,
    "TestLessThanOrEqualAddress": TestLessThanOrEqualAddressSchema,
    "U256InvModN": U256InvModNSchema,
    "Uint256DivMod": Uint256DivModSchema,
    "Uint256SquareRoot": Uint256SquareRootSchema,
    "Uint512DivModByUint256": Uint512DivModByUint256Schema,
    "WideMul128": WideMul128Schema,
}


class HintSchema(Schema):
    def load(self, data, *args, **kwargs) -> Hint:
        if not isinstance(data, dict) or len(data) != 1:
            raise ValidationError("Hint must be a dict with a single key.")

        key = next(iter(data))

        if key not in HINT_TYPE_SCHEMAS_MAPPING:
            raise ValidationError(f"Unknown Hint type: {key}")

        schema = HINT_TYPE_SCHEMAS_MAPPING[key]()

        return schema.load(data)

    def dump(self, obj, *args, **kwargs):
        if not isinstance(obj, dict) or len(obj) != 1:
            raise ValidationError("Hint must be a dict with a single key.")

        key = next(iter(obj))

        if key not in HINT_TYPE_SCHEMAS_MAPPING:
            raise ValidationError(f"Invalid value provided for Hint type: {key}")

        return {key: HINT_TYPE_SCHEMAS_MAPPING[key].dump(obj[key])}



import pytest

from starknet_py.net.models.chains import chain_from_network


def test_no_chain_for_custom_network():
    with pytest.raises(
        ValueError, match="Chain is required when not using predefined networks."
    ):
        chain_from_network("some_custom_url")



"""
The purpose of this file is to test serialization for complex abi.
"""

import json
from typing import NamedTuple

from starknet_py.abi.v0 import AbiParser as AbiParserV0
from starknet_py.abi.v1 import AbiParser as AbiParserV1
from starknet_py.abi.v2 import AbiParser as AbiParserV2
from starknet_py.cairo.felt import encode_shortstring
from starknet_py.serialization.factory import (
    serializer_for_event,
    serializer_for_function,
)
from starknet_py.serialization.tuple_dataclass import TupleDataclass
from starknet_py.tests.e2e.fixtures.abi_v1_structures import event_abi_v1
from starknet_py.tests.e2e.fixtures.constants import CAIRO_0_CONTRACTS_ABI_DIR
from starknet_py.tests.e2e.fixtures.misc import (
    ContractVersion,
    load_contract,
    read_contract,
)

dog = {"name": encode_shortstring("Cooper"), "species": encode_shortstring("dog")}
dog_serialized = [dog["name"], dog["species"]]

cat = {"name": encode_shortstring("Bob"), "species": encode_shortstring("cat")}
cat_serialized = [cat["name"], cat["species"]]

bird = {
    "name": encode_shortstring("Yacub"),
    "species": encode_shortstring("Parus major"),
}
bird_serialized = [bird["name"], bird["species"]]

ceo = {"name": encode_shortstring("Scrooge McDuck"), "pets": (dog, cat, bird)}
ceo_serialized = [ceo["name"], *dog_serialized, *cat_serialized, *bird_serialized]

company = {
    "id": 2**254 + 1234567890123457890,
    "name": encode_shortstring("McDuck Enterprises"),
    "address": encode_shortstring("Duckburg"),
    "owner": ceo,
    "company_structure": (1, (2, (3, 4, 5), 6, (7, 8), 9, (10,))),
}
company_serialized = [
    company["id"] % 2**128,
    company["id"] // 2**128,
    company["name"],
    company["address"],
    *ceo_serialized,
    *range(1, 11),
]


class School(NamedTuple):
    location: int
    name: int


class Education(NamedTuple):
    school: School
    level: int


university = School(
    location=encode_shortstring("Calisota"), name=encode_shortstring("Duckiversity")
)
university_serialized = [university.location, university.name]

education = Education(university, encode_shortstring("PhD"))
education_serialized = [*university_serialized, education.level]

person_gyro = {
    "name": encode_shortstring("Gyro Gearloose"),
    "education": education,
    "occupation": TupleDataclass.from_dict(
        {"company": company, "position": encode_shortstring("inventor")}
    ),
    "pets": (cat, dog),
}
person_gyro_serialized = [
    person_gyro["name"],
    *education_serialized,
    *company_serialized,
    person_gyro["occupation"].position,
    *cat_serialized,
    *dog_serialized,
]

person_donald = {
    "name": encode_shortstring("Donald Duck"),
    "education": education,
    "occupation": TupleDataclass.from_dict(
        {"company": company, "position": encode_shortstring("dunno")}
    ),
    "pets": (dog, dog),
}
person_donald_serialized = [
    person_donald["name"],
    *education_serialized,
    *company_serialized,
    person_donald["occupation"].position,
    *dog_serialized,
    *dog_serialized,
]

abi = json.loads(
    read_contract("complex_contract_abi.json", directory=CAIRO_0_CONTRACTS_ABI_DIR)
)
parsed_abi = AbiParserV0(abi).parse()

parsed_abi_v1 = AbiParserV1(event_abi_v1).parse()


def test_fn_serialization():
    expected_serialized = [
        2,
        *person_donald_serialized,
        *person_gyro_serialized,
        *company_serialized,
    ]
    serializer = serializer_for_function(parsed_abi.functions["hire"])

    assert expected_serialized == serializer.serialize(
        [person_donald, person_gyro], company
    )
    assert ([person_donald, person_gyro], company) == serializer.deserialize(
        expected_serialized
    )


def test_event_serialization_v0():
    expected_serialized = [
        *person_gyro_serialized,
        *company_serialized,
    ]
    serializer = serializer_for_event(parsed_abi.events["PersonHired"])
    payload = {"person": person_gyro, "company": company}

    assert expected_serialized == serializer.serialize(payload)
    assert TupleDataclass.from_dict(payload) == serializer.deserialize(
        expected_serialized
    )


def test_event_serialization_v1():
    serializer = serializer_for_event(parsed_abi_v1.events["Approval"])

    approval = {"owner": 1, "spender": 2, "value": 3}
    serialized = serializer.serialize(approval)
    assert serialized == [1, 2, 3, 0]

    assert serializer.deserialize(serialized).as_dict() == approval


def test_event_serialization_v2():
    abi_v2 = json.loads(
        load_contract(
            contract_name="NewSyntaxTestContract", version=ContractVersion.V2
        )["sierra"]
    )["abi"]

    parsed_abi_v2 = AbiParserV2(abi_v2).parse()
    serializer = serializer_for_event(
        parsed_abi_v2.events[
            "contracts_v2::new_syntax_test_contract::"
            "NewSyntaxTestContract::CounterIncreased"
        ]
    )

    serialized = serializer.serialize({"amount": 5})
    assert serialized == [5]

    assert serializer.deserialize(serialized).as_dict() == {"amount": 5}



import functools
from typing import List, Optional, Sequence

from Crypto.Hash import keccak
from crypto_cpp_py.cpp_bindings import (
    ECSignature,
    cpp_get_public_key,
    cpp_hash,
    cpp_sign,
    cpp_verify,
)

from starknet_py.common import int_from_bytes
from starknet_py.constants import EC_ORDER

MASK_250 = 2**250 - 1
HEX_PREFIX = "0x"


def _starknet_keccak(data: bytes) -> int:
    """
    A variant of eth-keccak that computes a value that fits in a Starknet field element.
    """
    k = keccak.new(digest_bits=256)
    k.update(data)
    return int_from_bytes(k.digest()) & MASK_250


def keccak256(data: bytes) -> int:
    k = keccak.new(digest_bits=256)
    k.update(data)
    return int_from_bytes(k.digest())


def pedersen_hash(left: int, right: int) -> int:
    """
    One of two hash functions (along with _starknet_keccak) used throughout Starknet.
    """
    return cpp_hash(left, right)


def compute_hash_on_elements(data: Sequence) -> int:
    """
    Computes a hash chain over the data, in the following order:
        h(h(h(h(0, data[0]), data[1]), ...), data[n-1]), n).

    The hash is initialized with 0 and ends with the data length appended.
    The length is appended in order to avoid collisions of the following kind:
    H([x,y,z]) = h(h(x,y),z) = H([w, z]) where w = h(x,y).
    """
    return functools.reduce(pedersen_hash, [*data, len(data)], 0)


def message_signature(
    msg_hash: int, priv_key: int, seed: Optional[int] = 32
) -> ECSignature:
    """
    Signs the message with private key.
    """
    return cpp_sign(msg_hash, priv_key, seed)


def verify_message_signature(
    msg_hash: int, signature: List[int], public_key: int
) -> bool:
    """
    Verifies ECDSA signature of a given message hash with a given public key.
    Returns true if public_key signs the message.
    """
    sig_r, sig_s = signature
    sig_w = pow(sig_s, -1, EC_ORDER)
    return cpp_verify(msg_hash=msg_hash, r=sig_r, w=sig_w, stark_key=public_key)


def private_to_stark_key(priv_key: int) -> int:
    """
    Deduces the public key given a private key.
    """
    return cpp_get_public_key(priv_key)


def encode_uint(value: int, bytes_length: int = 32) -> bytes:
    return value.to_bytes(bytes_length, byteorder="big")


def encode_uint_list(data: List[int]) -> bytes:
    return b"".join(encode_uint(x) for x in data)


def get_bytes_length(value: int) -> int:
    return (value.bit_length() + 7) // 8



import pytest

from starknet_py.net.client import Client
from starknet_py.net.client_models import PriceUnit
from starknet_py.tests.e2e.fixtures.accounts import mint_token_on_devnet
from starknet_py.tests.e2e.utils import _get_random_private_key_unsafe


@pytest.mark.asyncio
async def test_deploy_prefunded_account(
    account_with_validate_deploy_class_hash: int,
    client: Client,
):
    # pylint: disable=import-outside-toplevel, too-many-locals, unused-variable
    full_node_client_fixture = client
    # docs: start
    from starknet_py.hash.address import compute_address
    from starknet_py.net.account.account import Account
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping
    from starknet_py.net.full_node_client import FullNodeClient
    from starknet_py.net.signer.key_pair import KeyPair

    # First, make sure to generate private key and salt
    # docs: end
    private_key = _get_random_private_key_unsafe()
    salt = 1
    class_hash = account_with_validate_deploy_class_hash
    # docs: start
    key_pair = KeyPair.from_private_key(private_key)

    # You can also generate a key pair
    key_pair_generated = KeyPair.generate()

    # Compute an address
    address = compute_address(
        salt=salt,
        class_hash=class_hash,  # class_hash of the Account declared on the Starknet
        constructor_calldata=[key_pair.public_key],
        deployer_address=0,
    )

    # Prefund the address (using the token bridge or by sending fee tokens to the computed address)
    # Make sure the tx has been accepted on L2 before proceeding

    # Define the client to be used to interact with Starknet
    client = FullNodeClient(node_url="https://your.node.url")
    # docs: end

    client = full_node_client_fixture
    client_url = client.url.replace("/rpc", "")
    await mint_token_on_devnet(client_url, address, int(1e24), PriceUnit.FRI.value)

    # docs: start

    # Use `Account.deploy_account_v3` static method to deploy an account
    account_deployment_result = await Account.deploy_account_v3(
        address=address,
        class_hash=class_hash,
        salt=salt,
        key_pair=key_pair,
        client=client,
        constructor_calldata=[key_pair.public_key],
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e6), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    # Wait for deployment transaction to be accepted
    await account_deployment_result.wait_for_acceptance()

    # From now on, account can be used as usual
    account = account_deployment_result.account
    # docs: end
    assert account.address == address



from unittest.mock import MagicMock, Mock

import pytest

from starknet_py.net.models import StarknetChainId
from starknet_py.net.models.transaction import DeclareV3, DeployAccountV3, InvokeV3
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner


@pytest.mark.parametrize(
    "transaction",
    [
        Mock(spec=InvokeV3, calculate_hash=MagicMock(return_value=111)),
        Mock(spec=DeployAccountV3, calculate_hash=MagicMock(return_value=222)),
        Mock(spec=DeclareV3, calculate_hash=MagicMock(return_value=333)),
    ],
)
def test_sign_transaction(transaction):
    signer = StarkCurveSigner(
        account_address=0x1,
        key_pair=KeyPair.from_private_key(0x1),
        chain_id=StarknetChainId.MAINNET,
    )

    signature = signer.sign_transaction(transaction)

    assert isinstance(signature, list)
    assert len(signature) == 2
    assert all(isinstance(i, int) for i in signature)
    assert all(i != 0 for i in signature)


def test_key_pair():
    key_pair = KeyPair(public_key="0x123", private_key="0x456")

    assert isinstance(key_pair.public_key, int)
    assert isinstance(key_pair.private_key, int)

    key_pair = KeyPair.from_private_key("0x789")

    assert isinstance(key_pair.public_key, int)
    assert isinstance(key_pair.private_key, int)



import warnings
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from aiohttp import ClientResponse, ClientSession

from starknet_py.constants import EXPECTED_RPC_VERSION
from starknet_py.net.client_errors import ClientError


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"


class HttpClient(ABC):
    def __init__(self, url, session: Optional[ClientSession] = None):
        self.url = url
        self.session = session

    async def request(
        self,
        address: str,
        http_method: HttpMethod,
        params: Optional[dict] = None,
        payload: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
    ):
        kwargs = {
            "address": address,
            "http_method": http_method,
            "params": params,
            "payload": payload,
        }
        if self.session:
            return await self._make_request(session=self.session, **kwargs)

        async with ClientSession() as session:
            return await self._make_request(session=session, **kwargs)

    async def _make_request(
        self,
        session: ClientSession,
        address: str,
        http_method: HttpMethod,
        params: dict,
        payload: dict,
    ) -> dict:
        # pylint: disable=too-many-arguments
        async with session.request(
            method=http_method.value, url=address, params=params, json=payload
        ) as request:
            await self.handle_request_error(request)
            return await request.json(content_type=None)

    @abstractmethod
    async def handle_request_error(self, request: ClientResponse):
        """
        Handle an errors returned by make_request
        """


class RpcHttpClient(HttpClient):
    def __init__(
        self,
        url,
        session: Optional[ClientSession] = None,
        method_prefix: str = "starknet",
    ):
        super().__init__(url, session)
        self.method_prefix = method_prefix
        self._is_spec_version_verified: bool = False

    async def call(self, method_name: str, params: Optional[dict] = None):
        await self._warn_if_incompatible_rpc_version()

        payload = {
            "jsonrpc": "2.0",
            "method": f"{self.method_prefix}_{method_name}",
            "id": 0,
            "params": params if params else [],
        }

        result = await self.request(
            http_method=HttpMethod.POST, address=self.url, payload=payload
        )

        if "result" not in result:
            self.handle_rpc_error(result)
        return result["result"]

    @staticmethod
    def handle_rpc_error(result: dict):
        if "error" not in result:
            raise ServerError(body=result)
        raise ClientError(
            code=result["error"]["code"],
            message=result["error"]["message"],
            data=result["error"].get("data"),
        )

    async def handle_request_error(self, request: ClientResponse):
        await basic_error_handle(request)

    async def _warn_if_incompatible_rpc_version(self):
        if not self._is_spec_version_verified:
            payload = {
                "jsonrpc": "2.0",
                "method": "starknet_specVersion",
                "id": 0,
            }

            res = await self.request(
                http_method=HttpMethod.POST, address=self.url, payload=payload
            )
            spec_version = res["result"]

            if spec_version != EXPECTED_RPC_VERSION:
                warnings.warn(
                    f"RPC node with the url {self.url} uses incompatible version {spec_version}. "
                    f"Expected version: {EXPECTED_RPC_VERSION}",
                    IncompatibleRPCVersionWarning,
                    stacklevel=4,
                )
            self._is_spec_version_verified = True


async def basic_error_handle(request: ClientResponse):
    if request.status >= 300:
        raise ClientError(code=str(request.status), message=await request.text())


class ServerError(Exception):
    def __init__(self, body: dict):
        self.message = "RPC request failed."
        self.body = body
        super().__init__(self.message)


class IncompatibleRPCVersionWarning(Warning):
    pass



import pytest
from lark import Token, Tree

from starknet_py.abi.v1.parser_transformer import ParserTransformer


def test_default_parser_transformer():
    with pytest.raises(TypeError, match="Unable to parse tree node of type wrong."):
        ParserTransformer({}).transform(Tree(data=Token("RULE", "wrong"), children=[]))



from marshmallow import EXCLUDE, fields, post_load
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from starknet_py.net.client_models import (
    BlockTransactionTrace,
    DeclareTransactionTrace,
    DeployAccountTransactionTrace,
    FunctionInvocation,
    InvokeTransactionTrace,
    L1HandlerTransactionTrace,
    OrderedEvent,
    OrderedMessage,
    RevertedFunctionInvocation,
    SimulatedTransaction,
)
from starknet_py.net.schemas.common import CallTypeField, EntryPointTypeField, Felt
from starknet_py.net.schemas.rpc.block import StateDiffSchema
from starknet_py.net.schemas.rpc.general import (
    EstimatedFeeSchema,
    ExecutionResourcesSchema,
    InnerCallExecutionResourcesSchema,
)
from starknet_py.utils.schema import Schema


class OrderedEventSchema(Schema):
    keys = fields.List(Felt(), data_key="keys", required=True)
    data = fields.List(Felt(), data_key="data", required=True)
    order = fields.Integer(data_key="order", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs):
        return OrderedEvent(**data)


class OrderedMessageSchema(Schema):
    l2_address = Felt(data_key="from_address", required=True)
    l1_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)
    order = fields.Integer(data_key="order", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> OrderedMessage:
        return OrderedMessage(**data)


class FunctionInvocationSchema(Schema):
    contract_address = Felt(data_key="contract_address", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    caller_address = Felt(data_key="caller_address", required=True)
    class_hash = Felt(data_key="class_hash", required=True)
    entry_point_type = EntryPointTypeField(data_key="entry_point_type", required=True)
    call_type = CallTypeField(data_key="call_type", required=True)
    result = fields.List(Felt(), data_key="result", required=True)
    # https://marshmallow.readthedocs.io/en/stable/nesting.html#nesting-a-schema-within-itself
    calls = fields.List(
        fields.Nested(
            lambda: FunctionInvocationSchema()  # pylint: disable=unnecessary-lambda
        ),
        data_key="calls",
        required=True,
    )
    events = fields.List(
        fields.Nested(OrderedEventSchema()), data_key="events", required=True
    )
    messages = fields.List(
        fields.Nested(OrderedMessageSchema()), data_key="messages", required=True
    )
    execution_resources = fields.Nested(
        InnerCallExecutionResourcesSchema(),
        data_key="execution_resources",
        required=True,
    )
    is_reverted = fields.Boolean(data_key="is_reverted", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> FunctionInvocation:
        return FunctionInvocation(**data)


class RevertedFunctionInvocationSchema(Schema):
    revert_reason = fields.String(data_key="revert_reason", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> RevertedFunctionInvocation:
        return RevertedFunctionInvocation(**data)


class ExecuteInvocationSchema(OneOfSchema):
    type_schemas = {
        "REVERTED": RevertedFunctionInvocationSchema(),
        "FUNCTION_INVOCATION": FunctionInvocationSchema(),
    }

    def get_data_type(self, data):
        if "revert_reason" in data:
            return "REVERTED"
        return "FUNCTION_INVOCATION"


class InvokeTransactionTraceSchema(Schema):
    execute_invocation = fields.Nested(
        ExecuteInvocationSchema(), data_key="execute_invocation", required=True
    )
    execution_resources = fields.Nested(
        ExecutionResourcesSchema(), data_key="execution_resources", required=True
    )
    validate_invocation = fields.Nested(
        FunctionInvocationSchema(), data_key="validate_invocation", load_default=None
    )
    fee_transfer_invocation = fields.Nested(
        FunctionInvocationSchema(),
        data_key="fee_transfer_invocation",
        load_default=None,
    )
    state_diff = fields.Nested(
        StateDiffSchema(), data_key="state_diff", load_default=None
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> InvokeTransactionTrace:
        return InvokeTransactionTrace(**data)


class DeclareTransactionTraceSchema(Schema):
    execution_resources = fields.Nested(
        ExecutionResourcesSchema(), data_key="execution_resources", required=True
    )
    validate_invocation = fields.Nested(
        FunctionInvocationSchema(), data_key="validate_invocation", load_default=None
    )
    fee_transfer_invocation = fields.Nested(
        FunctionInvocationSchema(),
        data_key="fee_transfer_invocation",
        load_default=None,
    )
    state_diff = fields.Nested(
        StateDiffSchema(), data_key="state_diff", load_default=None
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionTrace:
        return DeclareTransactionTrace(**data)


class DeployAccountTransactionTraceSchema(Schema):
    constructor_invocation = fields.Nested(
        FunctionInvocationSchema(), data_key="constructor_invocation", required=True
    )
    execution_resources = fields.Nested(
        ExecutionResourcesSchema(), data_key="execution_resources", required=True
    )
    validate_invocation = fields.Nested(
        FunctionInvocationSchema(), data_key="validate_invocation", load_default=None
    )
    fee_transfer_invocation = fields.Nested(
        FunctionInvocationSchema(),
        data_key="fee_transfer_invocation",
        load_default=None,
    )
    state_diff = fields.Nested(
        StateDiffSchema(), data_key="state_diff", load_default=None
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployAccountTransactionTrace:
        return DeployAccountTransactionTrace(**data)


class L1HandlerTransactionTraceSchema(Schema):
    execution_resources = fields.Nested(
        ExecutionResourcesSchema(), data_key="execution_resources", required=True
    )

    function_invocation = fields.Nested(
        FunctionInvocationSchema(), data_key="function_invocation", required=True
    )
    state_diff = fields.Nested(
        StateDiffSchema(), data_key="state_diff", load_default=None
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> L1HandlerTransactionTrace:
        return L1HandlerTransactionTrace(**data)


class TransactionTraceSchema(OneOfSchema):
    type_field = "type"

    type_schemas = {
        "INVOKE": InvokeTransactionTraceSchema(),
        "DECLARE": DeclareTransactionTraceSchema(),
        "DEPLOY_ACCOUNT": DeployAccountTransactionTraceSchema(),
        "L1_HANDLER": L1HandlerTransactionTraceSchema(),
    }


class SimulatedTransactionSchema(Schema):
    # `unknown=EXCLUDE` in order to skip `type=...` field we don't want
    transaction_trace = fields.Nested(
        TransactionTraceSchema(),
        data_key="transaction_trace",
        required=True,
        unknown=EXCLUDE,
    )
    fee_estimation = fields.Nested(
        EstimatedFeeSchema(), data_key="fee_estimation", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> SimulatedTransaction:
        return SimulatedTransaction(**data)


class BlockTransactionTraceSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)
    # `unknown=EXCLUDE` in order to skip `type=...` field we don't want
    trace_root = fields.Nested(
        TransactionTraceSchema(), data_key="trace_root", required=True, unknown=EXCLUDE
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> BlockTransactionTrace:
        return BlockTransactionTrace(**data)



import pytest


@pytest.mark.asyncio
async def test_set_and_increase_time(devnet_client):
    time = 3384617820
    increase_value, error_margin = 100, 10

    await devnet_client.set_time(time, generate_block=True)
    block = await devnet_client.get_block(block_number="latest")

    assert block.timestamp == time

    await devnet_client.increase_time(increase_value)
    block = await devnet_client.get_block(block_number="latest")

    assert (
        time + increase_value <= block.timestamp <= time + increase_value + error_margin
    )


@pytest.mark.asyncio
async def test_set_time_with_generate_block(devnet_client):
    time = 3384617820

    block_number_before = await devnet_client.get_block_number()
    await devnet_client.set_time(time, generate_block=True)
    block_number_after = await devnet_client.get_block_number()

    assert block_number_after == block_number_before + 1


@pytest.mark.asyncio
async def test_set_time_without_generate_block(devnet_client):
    time = 3384617820

    block_number_before = await devnet_client.get_block_number()
    await devnet_client.set_time(time, generate_block=False)
    block_number_after = await devnet_client.get_block_number()

    assert block_number_after == block_number_before



from starknet_py.common import create_sierra_compiled_contract
from starknet_py.net.client_models import SierraCompiledContract


def test_create_new_compiled_contract(sierra_minimal_compiled_contract_and_class_hash):
    compiled_contract, _ = sierra_minimal_compiled_contract_and_class_hash

    contract = create_sierra_compiled_contract(compiled_contract)

    assert isinstance(contract, SierraCompiledContract)
    assert contract.abi is not None
